#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import numpy as np
import numba
import pytest

from opendsm.common.utils import (
    np_clip,
    OoM,
    RoundToSigFigs,
    t_stat,
    unc_factor,
    median_absolute_deviation,
    fast_std,
)


def test_np_clip():
    # Test case 1: Test with a scalar input
    a = 5
    a_min = 0
    a_max = 10
    result = np_clip(a, a_min, a_max)(a, a_min, a_max)
    expected = 5
    assert np.allclose(result, expected)

    # Test case 2: Test with an array input
    a = np.array([1, 2, 3, 4, 5])
    a_min = 2
    a_max = 4
    result = np_clip(a, a_min, a_max)(a, a_min, a_max)
    expected = np.array([2, 2, 3, 4, 4])
    assert np.allclose(result, expected)

    # Test case 3: Test with NaN values
    """
    We use the ~ operator to invert the boolean mask created by np.isnan(a), which replaces the NaN values with False. 
    We then use this mask to select the non-NaN values from the input array and the expected output array, and compare 
    them using np.allclose. This is to handle the issue where np.allclose returns False even though the result is the same as expected.
    """
    a = np.array([1, 2, np.nan, 4, 5])
    a_min = 2
    a_max = 4
    mask = ~np.isnan(a)
    result = np_clip(a[mask], a_min, a_max)(a[mask], a_min, a_max)
    expected = np.array([2, 2, np.nan, 4, 4])[mask]
    print(result, expected)
    assert np.allclose(result, expected)

    # Test case 4: Test with a_min > a_max (should raise ValueError)
    a = np.array([1, 2, 3, 4, 5])
    a_min = 4
    a_max = 2
    try:
        np_clip(a, a_min, a_max)(a, a_min, a_max)
    except ValueError as e:
        assert str(e) == "a_min must be less than or equal to a_max"


def test_OoM():
    # Test case 1: Test with a scalar input - should give an error as the declaration must have an array input
    x = 5000
    with pytest.raises(Exception) as e:
        OoM(x)
    assert e.type in [
        numba.core.errors.TypingError,
        TypeError,
    ]  # will depend whether using JIT

    # Test case 2: Test with an array input
    x = np.array([100, 1000, 10000, 100000])
    result = OoM(x)
    expected = np.array([2, 3, 4, 5])
    assert np.allclose(result, expected)

    # Test case 4: Test with a ceiling rounding method
    x = np.array([99, 999, 9999, 99999])
    result = OoM(x, method="ceil")
    expected = np.array([2, 3, 4, 5])
    assert np.allclose(result, expected)

    # Test case 4: Test with a floor rounding method
    x = np.array([101, 1001, 10001, 100001])
    result = OoM(x, method="floor")
    expected = np.array([2, 3, 4, 5])
    assert np.allclose(result, expected)

    # Test case 5: Test with an exact rounding method
    x = np.array([101, 1001, 10001, 100001])
    result = OoM(x, method="exact")
    expected = np.array([2, 3, 4, 5])
    assert np.allclose(result, expected)

    # Test case 6: Test with a non-integer input
    x = [1234.5678]
    result = OoM(x)
    expected = 3
    assert result == expected


def test_RoundToSigFigs():
    # Test case 1: Test with a scalar input
    x = [1234.5678]
    p = 3
    result = RoundToSigFigs(x, p)
    expected = 1230
    assert result == expected

    # Test case 2: Test with an array input
    x = np.array([1234.5678, 5678.1234, 0.0123456])
    p = 4
    result = RoundToSigFigs(x, p)
    expected = np.array([1.235e03, 5.680e03, 1.235e-02])
    assert np.allclose(result, expected)

    # Test case 3: Test with a zero input
    x = [0]
    p = 3
    result = RoundToSigFigs(x, p)
    expected = 0
    assert result == expected

    # Test case 4: Test with a negative input
    x = [-1234.5678]
    p = 3
    result = RoundToSigFigs(x, p)
    expected = -1230
    assert result == expected


def test_t_stat():
    # Test case 1: Test with a two-tailed test
    alpha = 0.05
    n = 10
    tail = 2
    result = t_stat(alpha, n, tail)
    expected = 2.262
    assert np.isclose(result, expected, rtol=1e-3)

    # Test case 2: Test with a one-tailed test
    alpha = 0.05
    n = 10
    tail = 1
    result = t_stat(alpha, n, tail)
    expected = 1.833
    assert np.isclose(result, expected, rtol=1e-3)

    # Test case 3: Test with a larger sample size
    alpha = 0.01
    n = 100
    tail = 2
    result = t_stat(alpha, n, tail)
    expected = 2.626
    assert np.isclose(result, expected, rtol=1e-3)

    # Test case 4: Test with a custom alpha value
    alpha = 0.10
    n = 10
    tail = 2
    result = t_stat(alpha, n, tail)
    expected = 1.833
    assert np.isclose(result, expected, rtol=1e-3)


def test_unc_factor():
    # Test case 1: Test with a confidence interval
    n = 10
    interval = "CI"
    alpha = 0.05
    result = unc_factor(n, interval, alpha)
    expected = 0.715
    assert np.isclose(result, expected, rtol=1e-3)

    # Test case 2: Test with a prediction interval
    n = 10
    interval = "PI"
    alpha = 0.05
    result = unc_factor(n, interval, alpha)
    expected = 2.977
    assert np.isclose(result, expected, rtol=1e-3)

    # Test case 3: Test with a larger sample size
    n = 100
    interval = "CI"
    alpha = 0.01
    result = unc_factor(n, interval, alpha)
    expected = 0.2626
    assert np.isclose(result, expected, rtol=1e-3)

    # Test case 4: Test with a custom alpha value
    n = 10
    interval = "PI"
    alpha = 0.10
    result = unc_factor(n, interval, alpha)
    expected = 2.412
    assert np.isclose(result, expected, rtol=1e-3)


def test_median_absolute_deviation():
    # Test case 1: Test with a small array
    x = np.array([1, 2, 3, 4, 5])
    result = median_absolute_deviation(x)
    expected = 1.4826
    assert np.isclose(result, expected, rtol=1e-3)

    # Test case 2: Test with a larger array
    x = np.random.normal(0, 1, 1000)
    result = median_absolute_deviation(x)
    expected = 1
    assert np.isclose(result, expected, rtol=1)

    # Test case 3: Test with an array of zeros
    x = np.zeros(10)
    result = median_absolute_deviation(x)
    expected = 0
    assert np.isclose(result, expected, rtol=1e-3)

    # Test case 4: Test with an array of ones
    x = np.ones(10)
    result = median_absolute_deviation(x)
    expected = 0
    assert np.isclose(result, expected, rtol=1e-3)


def test_fast_std():
    # Test case 1: Test with no weights and no mean
    x = np.array([1, 2, 3, 4, 5])
    result = fast_std(x)
    expected = 1.4142
    assert np.isclose(result, expected, rtol=1e-3)

    # Test case 2: Test with custom weights and no mean
    x = np.array([1, 2, 3, 4, 5])
    weights = np.array([0.1, 0.2, 0.3, 0.3, 0.1])
    result = fast_std(x, weights)
    expected = 1.270
    assert np.isclose(result, expected, rtol=1e-3)

    # Test case 3: Test with custom weights and custom mean
    x = np.array([1, 2, 3, 4, 5])
    weights = np.array([0.1, 0.2, 0.3, 0.2, 0.2])
    mean = 3
    result = fast_std(x, weights, mean)
    expected = 1.414
    assert np.isclose(result, expected, rtol=1e-3)

    # Test case 4: Test with a small array
    x = np.array([1])
    result = fast_std(x)
    expected = 0
    assert np.isclose(result, expected, rtol=1e-3)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import pytest
import numpy as np

from opendsm.eemeter.models.daily.optimize_results import (
    get_k,
    reduce_model,
    acf,
    OptimizedResult,
)


from opendsm.eemeter.models.daily.parameters import ModelCoefficients
from opendsm.eemeter.models.daily.utilities.settings import DailySettings as Settings


def test_get_k():
    # Test case 1: Test when all values are within the bounds
    X = [60, 0.5, 80, 0.3]
    T_min_seg = 50
    T_max_seg = 90
    assert get_k(X, T_min_seg, T_max_seg) == [70.0, 10.0, 74.0, 6.0]

    # Test case 2: Test when hdd_bp is greater than T_max_seg
    X = [100, 0.5, 80, 0.3]
    T_min_seg = 50
    T_max_seg = 90
    assert get_k(X, T_min_seg, T_max_seg) == [100, 0.0, 86.0, -6.0]

    # Test case 3: Test when cdd_bp is less than T_min_seg
    X = [60, 0.5, 20, 0.3]
    T_min_seg = 50
    T_max_seg = 90
    assert get_k(X, T_min_seg, T_max_seg) == [40.0, -20.0, 20, 0.0]

    # Test case 4: Test when both hdd_k and cdd_k are zero
    X = [100, 0.5, 20, 0.3]
    T_min_seg = 50
    T_max_seg = 90
    assert get_k(X, T_min_seg, T_max_seg) == [20, 0.0, 20, 0.0]


@pytest.mark.parametrize(
    "hdd_bp, hdd_beta, pct_hdd_k, cdd_bp, cdd_beta, pct_cdd_k, intercept, T_min, T_max, T_min_seg, T_max_seg, model_key, expected_coef_id, expected_x",
    [
        # Test case 1
        (
            10,
            20,
            30,
            40,
            50,
            60,
            70,
            0,
            100,
            20,
            80,
            "hdd_tidd_cdd_smooth",
            ["hdd_bp", "hdd_beta", "hdd_k", "cdd_bp", "cdd_beta", "cdd_k", "intercept"],
            [10, 20, 30, 40, 50, 60, 70],
        ),
        # Test case 2
        (
            10,
            20,
            0,
            40,
            50,
            60,
            70,
            0,
            100,
            20,
            80,
            "hdd_tidd_cdd_smooth",
            ["hdd_bp", "hdd_beta", "hdd_k", "cdd_bp", "cdd_beta", "cdd_k", "intercept"],
            [10, 20, 0, 40, 50, 60, 70],
        ),
        # Test case 3
        (
            10,
            0,
            30,
            40,
            50,
            60,
            70,
            0,
            100,
            20,
            80,
            "hdd_tidd_cdd_smooth",
            ["c_hdd_bp", "c_hdd_beta", "c_hdd_k", "intercept"],
            [20.0, 50.0, 20.0, 70.0],
        ),
        # Test case 4
        (
            10,
            20,
            0,
            40,
            50,
            0,
            70,
            0,
            100,
            20,
            80,
            "hdd_tidd_cdd_smooth",
            ["hdd_bp", "hdd_beta", "cdd_bp", "cdd_beta", "intercept"],
            [10, 20, 40, 50, 70],
        ),
        # Test case 5
        (
            10,
            0,
            0,
            40,
            0,
            0,
            70,
            0,
            100,
            20,
            80,
            "hdd_tidd_cdd_smooth",
            ["intercept"],
            [70],
        ),
    ],
)
def test_reduce_model(
    hdd_bp,
    hdd_beta,
    pct_hdd_k,
    cdd_bp,
    cdd_beta,
    pct_cdd_k,
    intercept,
    T_min,
    T_max,
    T_min_seg,
    T_max_seg,
    model_key,
    expected_coef_id,
    expected_x,
):
    coef_id, x = reduce_model(
        hdd_bp,
        hdd_beta,
        pct_hdd_k,
        cdd_bp,
        cdd_beta,
        pct_cdd_k,
        intercept,
        T_min,
        T_max,
        T_min_seg,
        T_max_seg,
        model_key,
    )

    assert coef_id == expected_coef_id
    assert np.allclose(x, expected_x)


def test_acf():
    # Test case 1: Test with a simple input array
    x = np.array([1, 2, 3, 4, 5])
    expected_output = np.array([1.0, 0.4, -0.1, -0.4])
    assert np.allclose(acf(x), expected_output)

    # Test case 3: Test with a moving mean and standard deviation
    x = np.array([1, 2, 3, 4, 5])
    expected_output = np.array([1.0, 1.0, 1.0, 1.0])
    assert np.allclose(acf(x, moving_mean_std=True), expected_output)

    # Test case 4: Test with a specific lag_n
    x = np.array([1, 2, 3, 4, 5])
    expected_output = np.array([1.0, 0.4])
    assert np.allclose(acf(x, lag_n=1), expected_output)


class TestOptimizeResult:
    @pytest.fixture
    def optimize_result(self):
        # create an instance of OptimizeResult for testing
        x = np.array([1, 2, 3, 4, 5, 6, 7])
        bnds = [[0, 10] * 7]
        coef_id = [
            "hdd_bp",
            "hdd_beta",
            "hdd_k",
            "cdd_bp",
            "cdd_beta",
            "cdd_k",
            "intercept",
        ]
        loss_alpha = 2.0
        C = np.array([1, 2, 3, 4, 5, 6, 7] * 2)
        T = np.array([1, 2, 3, 4, 5, 6, 7] * 2)
        model = np.array([1, 2, 3, 4, 5, 6, 7])
        resid = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7])
        weight = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7])
        settings = Settings()
        jac = None
        mean_loss = 0.0
        TSS = 0.0
        success = True
        message = "Optimization terminated successfully."
        nfev = 10
        time_elapsed = 1.0
        return OptimizedResult(
            x,
            bnds,
            coef_id,
            loss_alpha,
            T,
            C,
            model,
            weight,
            resid,
            jac,
            mean_loss,
            TSS,
            success,
            message,
            nfev,
            time_elapsed,
            settings,
        )

    def test_named_coeffs(self, optimize_result):
        # test that named_coeffs is an instance of ModelCoefficients
        assert isinstance(optimize_result.named_coeffs, ModelCoefficients)

    def test_prediction_uncertainty(self, optimize_result):
        # test that _prediction_uncertainty sets f_unc correctly
        optimize_result._prediction_uncertainty()
        assert optimize_result.f_unc == pytest.approx(2.1556496051287013)

    def test_set_model_key(self, optimize_result):
        # test that _set_model_key sets model_key and model_name correctly
        optimize_result._set_model_key()
        assert optimize_result.model_key == "c_hdd_tidd"
        assert optimize_result.model_name == "cdd_tidd"

    def test_refine_model(self, optimize_result):
        # test that _refine_model sets coef_id and x correctly
        optimize_result._refine_model()
        assert optimize_result.coef_id == ["c_hdd_bp", "c_hdd_beta", "intercept"]
        assert optimize_result.x == pytest.approx(np.array([4, 5, 7]))

    def test_eval(self, optimize_result):
        # test that eval returns the correct values
        T = np.array([1, 2, 3, 4, 5])
        model, f_unc, hdd_load, cdd_load = optimize_result.eval(T)
        assert model == pytest.approx(np.array([7, 7, 7, 7, 12]))
        assert f_unc == pytest.approx(
            np.array([2.15564961, 2.15564961, 2.15564961, 2.15564961, 2.15564961])
        )
        assert hdd_load == pytest.approx(np.array([0, 0, 0, 0, 0]))
        assert cdd_load == pytest.approx(np.array([0, 0, 0, 0, 5]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import numpy as np
import pandas as pd

from opendsm.eemeter.models.daily.utilities.ellipsoid_test import (
    ellipsoid_intersection_test,
    ellipsoid_K_function,
    robust_confidence_ellipse,
    ellipsoid_split_filter,
)


def test_ellipsoid_intersection_test():
    # test case 1: ellipsoids intersect
    mu_A = np.array([0, 0])
    mu_B = np.array([1, 1])
    cov_A = np.array([[1, 0], [0, 1]])
    cov_B = np.array([[1, 0], [0, 1]])
    assert ellipsoid_intersection_test(mu_A, mu_B, cov_A, cov_B) == True

    # test case 2: ellipsoids do not intersect
    mu_A = np.array([0, 0])
    mu_B = np.array([3, 3])
    cov_A = np.array([[1, 0], [0, 1]])
    cov_B = np.array([[1, 0], [0, 1]])
    assert ellipsoid_intersection_test(mu_A, mu_B, cov_A, cov_B) == False

    # test case 3: ellipsoids intersect at a single point
    mu_A = np.array([0, 0])
    mu_B = np.array([1, 0])
    cov_A = np.array([[1, 0], [0, 1]])
    cov_B = np.array([[1, 0], [0, 1]])
    assert ellipsoid_intersection_test(mu_A, mu_B, cov_A, cov_B) == True


def test_ellipsoid_K_function():
    # test case 1: ss = 0.5 -> doesn't match?
    ss = 0.5
    lambdas = np.array([1, 2, 3])
    v_squared = np.array([1, 2, 3])
    assert np.isclose(ellipsoid_K_function(ss, lambdas, v_squared), 0.0417, atol=1e-3)

    # test case 2: ss = 0
    ss = 0
    lambdas = np.array([1, 2])
    v_squared = np.array([1, 2])
    assert np.isclose(ellipsoid_K_function(ss, lambdas, v_squared), 1)

    # test case 3: ss = 1
    ss = 1
    lambdas = np.array([1, 2])
    v_squared = np.array([1, 2])
    assert np.isclose(ellipsoid_K_function(ss, lambdas, v_squared), 1)


def test_robust_confidence_ellipse():
    # these answers should remain constant
    def assert_close(mu, cov, a, b, phi):
        assert np.allclose(mu, np.array([10.4, 18.1]), rtol=0.1)
        assert np.allclose(cov, np.array([[27.2, -17], [-17, 28.4]]), rtol=0.1)
        assert np.isclose(a, 3.3, rtol=0.1)
        assert np.isclose(b, 6.7, rtol=0.1)
        assert np.isclose(phi, -2.4, rtol=0.1)

    # test case 1: no outliers
    # fmt: off
    x = np.array([ 4.9,  5.2,  6.4,  6.8,  8.6,  8.6, 10.3, 10.4, 10.9, 11.4, 12.2, 12.9, 15.1, 15.3, 17.1])
    y = np.array([24.5, 23.5, 13.8, 25.5, 13.6, 16.4, 15.4, 22.6,  8.9, 21.0, 13.6, 23.2, 21.7, 13.7, 14.1])
    # fmt: on
    mu, cov, a, b, phi = robust_confidence_ellipse(x, y)
    assert_close(mu, cov, a, b, phi)

    # test case 2: with outliers
    x = np.append(x, [8.7, 9.9, 10.5, 13, 13.4])
    y = np.append(y, [34.7, 4.4, 35.3, 5.9, 28.6])
    mu, cov, a, b, phi = robust_confidence_ellipse(x, y)
    assert_close(mu, cov, a, b, phi)


def test_ellipsoid_split_filter():
    # TODO : Add more test cases which contain all seasons and weekday/weekend
    # Test case 1: Test with a small dataset
    meter = pd.DataFrame(
        {
            "season": [
                "summer",
                "summer",
                "summer",
                "shoulder",
                "shoulder",
                "shoulder",
                "winter",
                "winter",
                "winter",
            ],
            "day_of_week": [1, 2, 3, 4, 5, 6, 7, 1, 2],
            "temperature": [20, 25, 30, 15, 20, 25, 10, 5, 0],
            "observed": [10, 20, 30, 15, 25, 35, 5, 10, 15],
        }
    )
    expected_output = {
        "summer": True,
        "shoulder": True,
        "winter": True,
        "weekday_weekend": True,
    }
    assert ellipsoid_split_filter(meter) == expected_output



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from datetime import datetime, timedelta

import numpy as np
import pandas as pd
import pytest
import pytz

from opendsm.eemeter.common.transform import (
    as_freq,
    clean_caltrack_billing_data,
    downsample_and_clean_caltrack_daily_data,
    clean_caltrack_billing_daily_data,
    day_counts,
    get_baseline_data,
    get_reporting_data,
    get_terms,
    remove_duplicates,
    NoBaselineDataError,
    NoReportingDataError,
    overwrite_partial_rows_with_nan,
    add_freq,
    trim,
    format_energy_data_for_caltrack,
    format_temperature_data_for_caltrack,
)


def test_as_freq_not_series(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    assert meter_data.shape == (27, 1)
    with pytest.raises(ValueError):
        as_freq(meter_data, freq="h")


def test_as_freq_hourly(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    assert meter_data.shape == (27, 1)
    as_hourly = as_freq(meter_data.value, freq="h")
    assert as_hourly.shape == (18961,)
    assert round(meter_data.value.sum(), 1) == round(as_hourly.sum(), 1) == 21290.2


def test_as_freq_daily(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    assert meter_data.shape == (27, 1)
    as_daily = as_freq(meter_data.value, freq="D")
    assert as_daily.shape == (792,)
    assert round(meter_data.value.sum(), 1) == round(as_daily.sum(), 1) == 21290.2


def test_as_freq_daily_all_nones_instantaneous(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    meter_data["value"] = np.nan
    assert meter_data.shape == (27, 1)
    as_daily = as_freq(meter_data.value, freq="D", series_type="instantaneous")
    assert as_daily.shape == (792,)
    assert round(meter_data.value.sum(), 1) == round(as_daily.sum(), 1) == 0


def test_as_freq_daily_all_nones(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    meter_data["value"] = np.nan
    assert meter_data.shape == (27, 1)
    as_daily = as_freq(meter_data.value, freq="D")
    assert as_daily.shape == (792,)
    assert round(meter_data.value.sum(), 1) == round(as_daily.sum(), 1) == 0


def test_as_freq_month_start(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    assert meter_data.shape == (27, 1)
    as_month_start = as_freq(meter_data.value, freq="MS")
    assert as_month_start.shape == (28,)
    assert round(meter_data.value.sum(), 1) == round(as_month_start.sum(), 1) == 21290.2


def test_as_freq_hourly_temperature(il_electricity_cdd_hdd_billing_monthly):
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    assert temperature_data.shape == (19417,)
    as_hourly = as_freq(temperature_data, freq="h", series_type="instantaneous")
    assert as_hourly.shape == (19417,)
    assert round(temperature_data.mean(), 1) == round(as_hourly.mean(), 1) == 54.6


def test_as_freq_daily_temperature(il_electricity_cdd_hdd_billing_monthly):
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    assert temperature_data.shape == (19417,)
    as_daily = as_freq(temperature_data, freq="D", series_type="instantaneous")
    assert as_daily.shape == (811,)
    assert abs(temperature_data.mean() - as_daily.mean()) <= 0.1


def test_as_freq_month_start_temperature(il_electricity_cdd_hdd_billing_monthly):
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    assert temperature_data.shape == (19417,)
    as_month_start = as_freq(temperature_data, freq="MS", series_type="instantaneous")
    assert as_month_start.shape == (29,)
    assert round(as_month_start.mean(), 1) == 53.4


def test_as_freq_daily_temperature_monthly(il_electricity_cdd_hdd_billing_monthly):
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    temperature_data = temperature_data.groupby(pd.Grouper(freq="MS")).mean()
    assert temperature_data.shape == (28,)
    as_daily = as_freq(temperature_data, freq="D", series_type="instantaneous")
    assert as_daily.shape == (824,)
    assert round(as_daily.mean(), 1) == 54.5


def test_as_freq_empty():
    meter_data = pd.DataFrame({"value": []})
    empty_meter_data = as_freq(meter_data.value, freq="h")
    assert empty_meter_data.empty


def test_as_freq_perserves_nulls(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    monthly_with_nulls = meter_data[meter_data.index.year != 2016].reindex(
        meter_data.index
    )
    daily_with_nulls = as_freq(monthly_with_nulls.value, freq="D")
    assert (
        round(monthly_with_nulls.value.sum(), 2)
        == round(daily_with_nulls.sum(), 2)
        == 11094.05
    )
    assert monthly_with_nulls.value.isnull().sum() == 13
    assert daily_with_nulls.isnull().sum() == 365


def test_day_counts(il_electricity_cdd_hdd_billing_monthly):
    data = il_electricity_cdd_hdd_billing_monthly["meter_data"].value
    counts = day_counts(data.index)
    assert counts.shape == (27,)
    assert counts.iloc[0] == 29.0
    assert pd.isnull(counts.iloc[-1])
    assert counts.sum() == 790.0


def test_day_counts_empty_series():
    index = pd.DatetimeIndex([])
    index.freq = None
    data = pd.Series([], index=index)
    counts = day_counts(data.index)
    assert counts.shape == (0,)


def test_get_baseline_data(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    baseline_data, warnings = get_baseline_data(meter_data)
    assert meter_data.shape == baseline_data.shape == (19417, 1)
    assert len(warnings) == 0


def test_get_baseline_data_with_timezones(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    baseline_data, warnings = get_baseline_data(
        meter_data.tz_convert("America/New_York")
    )
    assert len(warnings) == 0
    baseline_data, warnings = get_baseline_data(
        meter_data.tz_convert("Australia/Sydney")
    )
    assert len(warnings) == 0


def test_get_baseline_data_with_end(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    blackout_start_date = il_electricity_cdd_hdd_hourly["blackout_start_date"]
    baseline_data, warnings = get_baseline_data(meter_data, end=blackout_start_date)
    assert meter_data.shape != baseline_data.shape == (8761, 1)
    assert len(warnings) == 0


def test_get_baseline_data_with_end_no_max_days(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    blackout_start_date = il_electricity_cdd_hdd_hourly["blackout_start_date"]
    baseline_data, warnings = get_baseline_data(
        meter_data, end=blackout_start_date, max_days=None
    )
    assert meter_data.shape != baseline_data.shape == (9595, 1)
    assert len(warnings) == 0


def test_get_baseline_data_empty(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    blackout_start_date = il_electricity_cdd_hdd_hourly["blackout_start_date"]
    with pytest.raises(NoBaselineDataError):
        get_baseline_data(meter_data, end=pd.Timestamp("2000").tz_localize("UTC"))


def test_get_baseline_data_start_gap(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    start = meter_data.index.min() - timedelta(days=1)
    baseline_data, warnings = get_baseline_data(meter_data, start=start, max_days=None)
    assert meter_data.shape == baseline_data.shape == (19417, 1)
    assert len(warnings) == 1
    warning = warnings[0]
    assert warning.qualified_name == "eemeter.get_baseline_data.gap_at_baseline_start"
    assert (
        warning.description
        == "Data does not have coverage at requested baseline start date."
    )
    assert warning.data == {
        "data_start": "2015-11-22T06:00:00+00:00",
        "requested_start": "2015-11-21T06:00:00+00:00",
    }


def test_get_baseline_data_end_gap(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    end = meter_data.index.max() + timedelta(days=1)
    baseline_data, warnings = get_baseline_data(meter_data, end=end, max_days=None)
    assert meter_data.shape == baseline_data.shape == (19417, 1)
    assert len(warnings) == 1
    warning = warnings[0]
    assert warning.qualified_name == "eemeter.get_baseline_data.gap_at_baseline_end"
    assert (
        warning.description
        == "Data does not have coverage at requested baseline end date."
    )
    assert warning.data == {
        "data_end": "2018-02-08T06:00:00+00:00",
        "requested_end": "2018-02-09T06:00:00+00:00",
    }


def test_get_baseline_data_with_overshoot(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    baseline_data, warnings = get_baseline_data(
        meter_data,
        end=datetime(2016, 11, 9, tzinfo=pytz.UTC),
        max_days=32,
        allow_billing_period_overshoot=True,
    )
    assert baseline_data.shape == (2, 1)
    assert round(baseline_data.value.sum(), 2) == 632.31
    assert len(warnings) == 0

    baseline_data, warnings = get_baseline_data(
        meter_data,
        end=datetime(2016, 11, 9, tzinfo=pytz.UTC),
        max_days=32,
        allow_billing_period_overshoot=False,
    )
    assert baseline_data.shape == (1, 1)
    assert round(baseline_data.value.sum(), 2) == 0
    assert len(warnings) == 0

    baseline_data, warnings = get_baseline_data(
        meter_data,
        end=datetime(2016, 11, 9, tzinfo=pytz.UTC),
        max_days=25,
        allow_billing_period_overshoot=True,
    )
    assert baseline_data.shape == (1, 1)
    assert round(baseline_data.value.sum(), 2) == 0
    assert len(warnings) == 0


def test_get_baseline_data_with_ignored_gap(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    baseline_data, warnings = get_baseline_data(
        meter_data,
        end=datetime(2016, 11, 9, tzinfo=pytz.UTC),
        max_days=45,
        ignore_billing_period_gap_for_day_count=True,
    )
    assert baseline_data.shape == (2, 1)
    assert round(baseline_data.value.sum(), 2) == 632.31
    assert len(warnings) == 0

    baseline_data, warnings = get_baseline_data(
        meter_data,
        end=datetime(2016, 11, 9, tzinfo=pytz.UTC),
        max_days=45,
        ignore_billing_period_gap_for_day_count=False,
    )
    assert baseline_data.shape == (1, 1)
    assert round(baseline_data.value.sum(), 2) == 0
    assert len(warnings) == 0

    baseline_data, warnings = get_baseline_data(
        meter_data,
        end=datetime(2016, 11, 9, tzinfo=pytz.UTC),
        max_days=25,
        ignore_billing_period_gap_for_day_count=True,
    )
    assert baseline_data.shape == (1, 1)
    assert round(baseline_data.value.sum(), 2) == 0
    assert len(warnings) == 0


def test_get_baseline_data_with_overshoot_and_ignored_gap(
    il_electricity_cdd_hdd_billing_monthly,
):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    baseline_data, warnings = get_baseline_data(
        meter_data,
        end=datetime(2016, 11, 9, tzinfo=pytz.UTC),
        max_days=25,
        allow_billing_period_overshoot=True,
        ignore_billing_period_gap_for_day_count=True,
    )
    assert baseline_data.shape == (2, 1)
    assert round(baseline_data.value.sum(), 2) == 632.31
    assert len(warnings) == 0

    baseline_data, warnings = get_baseline_data(
        meter_data,
        end=datetime(2016, 11, 9, tzinfo=pytz.UTC),
        max_days=25,
        allow_billing_period_overshoot=False,
        ignore_billing_period_gap_for_day_count=False,
    )
    assert baseline_data.shape == (1, 1)
    assert round(baseline_data.value.sum(), 2) == 0
    assert len(warnings) == 0


def test_get_baseline_data_n_days_billing_period_overshoot(
    il_electricity_cdd_hdd_billing_monthly,
):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    baseline_data, warnings = get_baseline_data(
        meter_data,
        end=datetime(2017, 11, 9, tzinfo=pytz.UTC),
        max_days=45,
        allow_billing_period_overshoot=True,
        n_days_billing_period_overshoot=45,
        ignore_billing_period_gap_for_day_count=True,
    )
    assert baseline_data.shape == (2, 1)
    assert round(baseline_data.value.sum(), 2) == 526.25
    assert len(warnings) == 0


def test_get_baseline_data_too_far_from_date(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    end_date = datetime(2020, 11, 9, tzinfo=pytz.UTC)
    max_days = 45
    baseline_data, warnings = get_baseline_data(
        meter_data,
        end=end_date,
        max_days=max_days,
        ignore_billing_period_gap_for_day_count=True,
    )
    assert baseline_data.shape == (2, 1)
    assert round(baseline_data.value.sum(), 2) == 1393.4
    assert len(warnings) == 0
    with pytest.raises(NoBaselineDataError):
        get_baseline_data(
            meter_data,
            end=end_date,
            max_days=max_days,
            n_days_billing_period_overshoot=45,
            ignore_billing_period_gap_for_day_count=True,
        )
    baseline_data, warnings = get_baseline_data(
        meter_data,
        end=end_date,
        max_days=max_days,
        allow_billing_period_overshoot=True,
        ignore_billing_period_gap_for_day_count=True,
    )
    assert baseline_data.shape == (3, 1)
    assert round(baseline_data.value.sum(), 2) == 2043.92
    assert len(warnings) == 0
    # Includes 3 data points because data at index -3 is closer to start target
    # then data at index -2
    start_target = baseline_data.index[-1] - timedelta(days=max_days)
    assert abs((baseline_data.index[0] - start_target).days) < abs(
        (baseline_data.index[1] - start_target).days
    )
    with pytest.raises(NoBaselineDataError):
        get_baseline_data(
            meter_data,
            end=end_date,
            max_days=max_days,
            allow_billing_period_overshoot=True,
            n_days_billing_period_overshoot=45,
            ignore_billing_period_gap_for_day_count=True,
        )


def test_get_reporting_data(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    reporting_data, warnings = get_reporting_data(meter_data)
    assert meter_data.shape == reporting_data.shape == (19417, 1)
    assert len(warnings) == 0


def test_get_reporting_data_with_timezones(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    reporting_data, warnings = get_reporting_data(
        meter_data.tz_convert("America/New_York")
    )
    assert len(warnings) == 0
    reporting_data, warnings = get_reporting_data(
        meter_data.tz_convert("Australia/Sydney")
    )
    assert len(warnings) == 0


def test_get_reporting_data_with_start(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    blackout_end_date = il_electricity_cdd_hdd_hourly["blackout_end_date"]
    reporting_data, warnings = get_reporting_data(meter_data, start=blackout_end_date)
    assert meter_data.shape != reporting_data.shape == (8761, 1)
    assert len(warnings) == 0


def test_get_reporting_data_with_start_no_max_days(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    blackout_end_date = il_electricity_cdd_hdd_hourly["blackout_end_date"]
    reporting_data, warnings = get_reporting_data(
        meter_data, start=blackout_end_date, max_days=None
    )
    assert meter_data.shape != reporting_data.shape == (9607, 1)
    assert len(warnings) == 0


def test_get_reporting_data_empty(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    blackout_end_date = il_electricity_cdd_hdd_hourly["blackout_end_date"]
    with pytest.raises(NoReportingDataError):
        get_reporting_data(meter_data, start=pd.Timestamp("2030").tz_localize("UTC"))


def test_get_reporting_data_start_gap(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    start = meter_data.index.min() - timedelta(days=1)
    reporting_data, warnings = get_reporting_data(
        meter_data, start=start, max_days=None
    )
    assert meter_data.shape == reporting_data.shape == (19417, 1)
    assert len(warnings) == 1
    warning = warnings[0]
    assert warning.qualified_name == "eemeter.get_reporting_data.gap_at_reporting_start"
    assert (
        warning.description
        == "Data does not have coverage at requested reporting start date."
    )
    assert warning.data == {
        "data_start": "2015-11-22T06:00:00+00:00",
        "requested_start": "2015-11-21T06:00:00+00:00",
    }


def test_get_reporting_data_end_gap(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    end = meter_data.index.max() + timedelta(days=1)
    reporting_data, warnings = get_reporting_data(meter_data, end=end, max_days=None)
    assert meter_data.shape == reporting_data.shape == (19417, 1)
    assert len(warnings) == 1
    warning = warnings[0]
    assert warning.qualified_name == "eemeter.get_reporting_data.gap_at_reporting_end"
    assert (
        warning.description
        == "Data does not have coverage at requested reporting end date."
    )
    assert warning.data == {
        "data_end": "2018-02-08T06:00:00+00:00",
        "requested_end": "2018-02-09T06:00:00+00:00",
    }


def test_get_reporting_data_with_overshoot(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    reporting_data, warnings = get_reporting_data(
        meter_data,
        start=datetime(2016, 9, 9, tzinfo=pytz.UTC),
        max_days=30,
        allow_billing_period_overshoot=True,
    )
    assert reporting_data.shape == (2, 1)
    assert round(reporting_data.value.sum(), 2) == 632.31
    assert len(warnings) == 0

    reporting_data, warnings = get_reporting_data(
        meter_data,
        start=datetime(2016, 9, 9, tzinfo=pytz.UTC),
        max_days=30,
        allow_billing_period_overshoot=False,
    )
    assert reporting_data.shape == (1, 1)
    assert round(reporting_data.value.sum(), 2) == 0
    assert len(warnings) == 0

    reporting_data, warnings = get_reporting_data(
        meter_data,
        start=datetime(2016, 9, 9, tzinfo=pytz.UTC),
        max_days=25,
        allow_billing_period_overshoot=True,
    )
    assert reporting_data.shape == (1, 1)
    assert round(reporting_data.value.sum(), 2) == 0
    assert len(warnings) == 0


def test_get_reporting_data_with_ignored_gap(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    reporting_data, warnings = get_reporting_data(
        meter_data,
        start=datetime(2016, 9, 9, tzinfo=pytz.UTC),
        max_days=45,
        ignore_billing_period_gap_for_day_count=True,
    )
    assert reporting_data.shape == (2, 1)
    assert round(reporting_data.value.sum(), 2) == 632.31
    assert len(warnings) == 0

    reporting_data, warnings = get_reporting_data(
        meter_data,
        start=datetime(2016, 9, 9, tzinfo=pytz.UTC),
        max_days=45,
        ignore_billing_period_gap_for_day_count=False,
    )
    assert reporting_data.shape == (1, 1)
    assert round(reporting_data.value.sum(), 2) == 0
    assert len(warnings) == 0

    reporting_data, warnings = get_reporting_data(
        meter_data,
        start=datetime(2016, 9, 9, tzinfo=pytz.UTC),
        max_days=25,
        ignore_billing_period_gap_for_day_count=True,
    )
    assert reporting_data.shape == (1, 1)
    assert round(reporting_data.value.sum(), 2) == 0
    assert len(warnings) == 0


def test_get_reporting_data_with_overshoot_and_ignored_gap(
    il_electricity_cdd_hdd_billing_monthly,
):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    reporting_data, warnings = get_reporting_data(
        meter_data,
        start=datetime(2016, 9, 9, tzinfo=pytz.UTC),
        max_days=25,
        allow_billing_period_overshoot=True,
        ignore_billing_period_gap_for_day_count=True,
    )
    assert reporting_data.shape == (2, 1)
    assert round(reporting_data.value.sum(), 2) == 632.31
    assert len(warnings) == 0

    reporting_data, warnings = get_reporting_data(
        meter_data,
        start=datetime(2016, 9, 9, tzinfo=pytz.UTC),
        max_days=25,
        allow_billing_period_overshoot=False,
        ignore_billing_period_gap_for_day_count=False,
    )
    assert reporting_data.shape == (1, 1)
    assert round(reporting_data.value.sum(), 2) == 0
    assert len(warnings) == 0


def test_get_terms_unrecognized_method(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]

    with pytest.raises(ValueError):
        get_terms(meter_data.index, term_lengths=[365], method="unrecognized")


def test_get_terms_unsorted_index(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]

    with pytest.raises(ValueError):
        get_terms(meter_data.index[::-1], term_lengths=[365])


def test_get_terms_bad_term_labels(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]

    with pytest.raises(ValueError):
        terms = get_terms(
            meter_data.index,
            term_lengths=[60, 60, 60],
            term_labels=["abc", "def"],  # too short
        )


def test_get_terms_default_term_labels(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]

    terms = get_terms(meter_data.index, term_lengths=[60, 60, 60])
    assert [t.label for t in terms] == ["term_001", "term_002", "term_003"]


def test_get_terms_custom_term_labels(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]

    terms = get_terms(
        meter_data.index, term_lengths=[60, 60, 60], term_labels=["abc", "def", "ghi"]
    )
    assert [t.label for t in terms] == ["abc", "def", "ghi"]


def test_get_terms_empty_index_input(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]

    terms = get_terms(meter_data.index[:0], term_lengths=[60, 60, 60])
    assert len(terms) == 0


def test_get_terms_strict(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]

    strict_terms = get_terms(
        meter_data.index,
        term_lengths=[365, 365],
        term_labels=["year1", "year2"],
        start=datetime(2016, 1, 15, tzinfo=pytz.UTC),
        method="strict",
    )

    assert len(strict_terms) == 2

    year1 = strict_terms[0]
    assert year1.label == "year1"
    assert year1.index.shape == (12,)
    assert (
        year1.target_start_date
        == pd.Timestamp("2016-01-15 00:00:00+0000", tz="UTC").to_pydatetime()
    )
    assert (
        year1.target_end_date
        == pd.Timestamp("2017-01-14 00:00:00+0000", tz="UTC").to_pydatetime()
    )
    assert year1.target_term_length_days == 365
    assert (
        year1.actual_start_date
        == year1.index[0]
        == pd.Timestamp("2016-01-22 06:00:00+0000", tz="UTC")
    )
    assert (
        year1.actual_end_date
        == year1.index[-1]
        == pd.Timestamp("2016-12-19 06:00:00+0000", tz="UTC")
    )
    assert year1.actual_term_length_days == 332
    assert year1.complete

    year2 = strict_terms[1]
    assert year2.index.shape == (13,)
    assert year2.label == "year2"
    assert year2.target_start_date == pd.Timestamp("2016-12-19 06:00:00+0000", tz="UTC")
    assert (
        year2.target_end_date
        == pd.Timestamp("2018-01-14 00:00:00+0000", tz="UTC").to_pydatetime()
    )
    assert year2.target_term_length_days == 365
    assert (
        year2.actual_start_date
        == year2.index[0]
        == pd.Timestamp("2016-12-19 06:00:00+00:00", tz="UTC")
    )
    assert (
        year2.actual_end_date
        == year2.index[-1]
        == pd.Timestamp("2017-12-22 06:00:00+0000", tz="UTC")
    )
    assert year2.actual_term_length_days == 368
    assert year2.complete


def test_get_terms_nearest(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    nearest_terms = get_terms(
        meter_data.index,
        term_lengths=[365, 365],
        term_labels=["year1", "year2"],
        start=datetime(2016, 1, 15, tzinfo=pytz.UTC),
        method="nearest",
    )

    assert len(nearest_terms) == 2

    year1 = nearest_terms[0]
    assert year1.label == "year1"
    assert year1.index.shape == (13,)
    assert year1.index[0] == pd.Timestamp("2016-01-22 06:00:00+0000", tz="UTC")
    assert year1.index[-1] == pd.Timestamp("2017-01-21 06:00:00+0000", tz="UTC")
    assert (
        year1.target_start_date
        == pd.Timestamp("2016-01-15 00:00:00+0000", tz="UTC").to_pydatetime()
    )
    assert year1.target_term_length_days == 365
    assert year1.actual_term_length_days == 365
    assert year1.complete

    year2 = nearest_terms[1]
    assert year2.label == "year2"
    assert year2.index.shape == (13,)
    assert year2.index[0] == pd.Timestamp("2017-01-21 06:00:00+0000", tz="UTC")
    assert year2.index[-1] == pd.Timestamp("2018-01-20 06:00:00+0000", tz="UTC")
    assert year2.target_start_date == pd.Timestamp("2017-01-21 06:00:00+0000", tz="UTC")
    assert year1.target_term_length_days == 365
    assert year2.actual_term_length_days == 364
    assert not year2.complete  # no remaining index

    # check completeness case with a shorter final term
    nearest_terms = get_terms(
        meter_data.index,
        term_lengths=[365, 340],
        term_labels=["year1", "year2"],
        start=datetime(2016, 1, 15, tzinfo=pytz.UTC),
        method="nearest",
    )
    year2 = nearest_terms[1]
    assert year2.label == "year2"
    assert year2.index.shape == (12,)
    assert year2.index[0] == pd.Timestamp("2017-01-21 06:00:00+0000", tz="UTC")
    assert year2.index[-1] == pd.Timestamp("2017-12-22 06:00:00+00:00", tz="UTC")
    assert year2.target_start_date == pd.Timestamp("2017-01-21 06:00:00+0000", tz="UTC")
    assert year2.target_term_length_days == 340
    assert year2.actual_term_length_days == 335
    assert year2.complete  # has remaining index


def test_term_repr(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]

    terms = get_terms(meter_data.index, term_lengths=[60, 60, 60])
    assert repr(terms[0]) == (
        "Term(label=term_001, target_term_length_days=60, actual_term_length_days=29,"
        " complete=True)"
    )


def test_remove_duplicates_df():
    index = pd.DatetimeIndex(["2017-01-01", "2017-01-02", "2017-01-02"])
    df = pd.DataFrame({"value": [1, 2, 3]}, index=index)
    assert df.shape == (3, 1)
    df_dedupe = remove_duplicates(df)
    assert df_dedupe.shape == (2, 1)
    assert list(df_dedupe.value) == [1, 2]


def test_remove_duplicates_series():
    index = pd.DatetimeIndex(["2017-01-01", "2017-01-02", "2017-01-02"])
    series = pd.Series([1, 2, 3], index=index)
    assert series.shape == (3,)
    series_dedupe = remove_duplicates(series)
    assert series_dedupe.shape == (2,)
    assert list(series_dedupe) == [1, 2]


def test_as_freq_hourly_to_daily(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]

    meter_data.iloc[-1, meter_data.columns.get_loc("value")] = np.nan
    assert meter_data.shape == (19417, 1)
    as_daily = as_freq(meter_data.value, freq="D")
    assert as_daily.shape == (811,)
    assert round(meter_data.value.sum(), 1) == round(as_daily.sum(), 1) == 21926.0


def test_as_freq_daily_to_daily(il_electricity_cdd_hdd_daily):
    meter_data = il_electricity_cdd_hdd_daily["meter_data"]
    assert meter_data.shape == (810, 1)
    as_daily = as_freq(meter_data.value, freq="D")
    assert as_daily.shape == (810,)
    assert round(meter_data.value.sum(), 1) == round(as_daily.sum(), 1) == 21925.8


def test_as_freq_hourly_to_daily_include_coverage(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    meter_data.iloc[-1, meter_data.columns.get_loc("value")] = np.nan
    assert meter_data.shape == (19417, 1)
    as_daily = as_freq(meter_data.value, freq="D", include_coverage=True)
    assert as_daily.shape == (811, 2)
    assert round(meter_data.value.sum(), 1) == round(as_daily.value.sum(), 1) == 21926.0


def test_clean_caltrack_billing_daily_data_billing(
    il_electricity_cdd_hdd_billing_monthly,
):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    cleaned_data = clean_caltrack_billing_daily_data(meter_data, "billing_monthly")
    assert cleaned_data.shape == (27, 1)
    pd.testing.assert_frame_equal(meter_data, cleaned_data)


def test_clean_caltrack_billing_daily_data_daily(il_electricity_cdd_hdd_daily):
    meter_data = il_electricity_cdd_hdd_daily["meter_data"]
    cleaned_data = clean_caltrack_billing_daily_data(meter_data, "daily")
    assert cleaned_data.shape == (810, 1)
    pd.testing.assert_frame_equal(meter_data, cleaned_data)


def test_clean_caltrack_billing_daily_data_daily_local_tz(il_electricity_cdd_hdd_daily):
    meter_data = il_electricity_cdd_hdd_daily["meter_data"]
    meter_data.index += timedelta(hours=6)
    meter_data = meter_data.tz_convert("America/Chicago")
    cleaned_data = clean_caltrack_billing_daily_data(meter_data, "daily")
    assert cleaned_data.shape == (810, 1)
    pd.testing.assert_frame_equal(meter_data, cleaned_data)


def test_clean_caltrack_billing_daily_data_hourly(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    cleaned_data = clean_caltrack_billing_daily_data(meter_data, "hourly")
    assert cleaned_data.shape == (811, 1)


def test_clean_caltrack_daily_data_hourly(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    cleaned_data = downsample_and_clean_caltrack_daily_data(meter_data)
    assert cleaned_data.shape == (811, 1)


def test_clean_caltrack_daily_data_hourly_local_tz(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    meter_data = meter_data.tz_convert("America/Chicago")
    cleaned_data = downsample_and_clean_caltrack_daily_data(meter_data)
    assert cleaned_data.shape == (810, 1)


def test_clean_caltrack_billing_data_estimated(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    meter_data["estimated"] = False
    estimated_col_index = meter_data.columns.get_loc("estimated")
    meter_data.iloc[:, estimated_col_index] = False
    meter_data.iloc[2, estimated_col_index] = True
    meter_data.iloc[5, estimated_col_index] = True
    meter_data.iloc[6, estimated_col_index] = True
    meter_data.iloc[10, estimated_col_index] = True

    cleaned_data = clean_caltrack_billing_data(meter_data, "billing_monthly")
    assert cleaned_data.dropna().shape[0] == cleaned_data.shape[0] - 2


def test_clean_caltrack_billing_data_uneven_datetimes(
    il_electricity_cdd_hdd_billing_monthly,
):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    too_short_meter_data = pd.concat(
        [
            meter_data,
            pd.DataFrame(
                data=[{"value": 100}],
                index=[datetime(2017, 1, 1, 6).replace(tzinfo=pytz.UTC)],
            ),
        ]
    ).sort_index()
    cleaned_data = clean_caltrack_billing_data(too_short_meter_data, "billing_monthly")
    assert cleaned_data.dropna().shape[0] == cleaned_data.shape[0] - 3

    too_long_meter_data = meter_data.drop(
        [datetime(2016, 12, 19, 6).replace(tzinfo=pytz.UTC)]
    )
    cleaned_data = clean_caltrack_billing_data(too_long_meter_data, "billing_monthly")

    too_long_meter_data = meter_data.drop(
        [
            datetime(2016, 12, 19, 6).replace(tzinfo=pytz.UTC),
            datetime(2017, 1, 21, 6).replace(tzinfo=pytz.UTC),
        ]
    )
    cleaned_data = clean_caltrack_billing_data(too_long_meter_data, "billing_bimonthly")
    assert cleaned_data.dropna().shape[0] == cleaned_data.shape[0] - 2
    assert cleaned_data.dropna().shape[0] == cleaned_data.shape[0] - 2

    pre_empty_meter_data = meter_data[:0]
    cleaned_data = clean_caltrack_billing_data(pre_empty_meter_data, "billing_monthly")
    assert cleaned_data.empty

    post_empty_meter_data = meter_data[:4].drop(
        [
            datetime(2015, 12, 21, 6).replace(tzinfo=pytz.UTC),
            datetime(2016, 1, 22, 6).replace(tzinfo=pytz.UTC),
        ]
    )
    assert not post_empty_meter_data["value"].dropna().empty
    cleaned_data = clean_caltrack_billing_data(post_empty_meter_data, "billing_monthly")
    assert cleaned_data.empty


def test_overwrite_partial_rows_with_nan(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    meter_data["other_column"] = meter_data["value"]
    meter_data.iloc[:3, meter_data.columns.get_loc("other_column")] = np.nan
    meter_data_nanned = overwrite_partial_rows_with_nan(meter_data)
    assert pd.isnull(meter_data_nanned["value"][:3]).all()


import pandas as pd


def test_add_freq(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]

    # make DateTimeIndex timezone-naive
    meter_data.index = meter_data.index.tz_localize(None)

    # infer frequency
    meter_data.index = add_freq(meter_data.index)
    assert meter_data.index.freq == "h"


def test_trim_two_dataframes(
    uk_electricity_hdd_only_hourly_sample_1, uk_electricity_hdd_only_hourly_sample_2
):
    df1 = uk_electricity_hdd_only_hourly_sample_1["meter_data"]
    df2 = uk_electricity_hdd_only_hourly_sample_2["meter_data"]

    df1_trimmed, df2_trimmed = trim(df1, df2)

    assert (
        df1.index[0] == df1.index.min()
        and df2.index[0] == df2.index.min()
        and df1.index[0] != df2.index[0]
    )

    assert (
        df1.index[-1] == df1.index.max()
        and df2.index[-1] == df2.index.max()
        and df1.index[-1] != df2.index[-1]
    )

    assert df1_trimmed.index[0] == df2_trimmed.index[0]
    assert df1_trimmed.index.min() == df2_trimmed.index.min()
    assert df1_trimmed.index[-1] == df2_trimmed.index[-1]
    assert df1_trimmed.index.max() == df2_trimmed.index.max()


def test_format_temperature_data_for_caltrack(il_electricity_cdd_hdd_hourly):
    temperature_data = il_electricity_cdd_hdd_hourly["temperature_data"]

    # temperature_data to pd.DateFrame
    temperature_data = pd.DataFrame(temperature_data)

    # flipping df
    temperature_data = temperature_data.reindex(index=temperature_data.index[::-1])

    # inserting new value of 0.04 at 09.34 22/11/2015
    new_start = pd.to_datetime("22/11/2015 09:34", dayfirst=True).tz_localize("UTC")
    temperature_data.loc[new_start] = [0.04]

    # rename column name to 'consumption'
    temperature_data.rename(columns={"value": "consumption"}, inplace=True)

    temperature_data_reformatted = format_temperature_data_for_caltrack(
        temperature_data
    )

    assert isinstance(temperature_data_reformatted, pd.Series)
    assert (
        temperature_data_reformatted.index[0] < temperature_data_reformatted.index[-1]
    )
    assert temperature_data_reformatted.index.freq == "h"
    assert temperature_data_reformatted.index.tzinfo is not None


def test_format_energy_data_for_caltrack_hourly(il_electricity_cdd_hdd_hourly):
    df = il_electricity_cdd_hdd_hourly["meter_data"]
    # flipping df
    df = df.reindex(index=df.index[::-1])

    # inserting new value of 0.04 at 09.34 22/11/2015
    new_start = pd.to_datetime("22/11/2015 09:34", dayfirst=True).tz_localize("UTC")
    df.loc[new_start] = [0.04]

    # rename column name to 'consumption'
    df.rename(columns={"value": "consumption"}, inplace=True)

    # df_flipped to pd.Series
    df = df.squeeze()

    df_reformatted = format_energy_data_for_caltrack(df, method="hourly")

    assert isinstance(df_reformatted, pd.DataFrame)
    assert df_reformatted.index[0] < df_reformatted.index[-1]
    assert df_reformatted.index.freq == "h"
    assert df_reformatted.columns[0] == "value"
    assert df_reformatted.index.tzinfo is not None
    assert len(df_reformatted.columns) == 1


def test_format_energy_data_for_caltrack_daily(il_electricity_cdd_hdd_daily):
    df = il_electricity_cdd_hdd_daily["meter_data"]
    # flipping df
    df = df.reindex(index=df.index[::-1])

    # inserting new value of 0.04 at 09.34 22/11/2015
    new_start = pd.to_datetime("22/11/2015 09:34", dayfirst=True).tz_localize("UTC")
    df.loc[new_start] = [0.04]

    # rename column name to 'consumption'
    df.rename(columns={"value": "consumption"}, inplace=True)

    # df_flipped to pd.Series
    df = df.squeeze()

    df_reformatted = format_energy_data_for_caltrack(df, method="daily")

    assert isinstance(df_reformatted, pd.DataFrame)
    assert df_reformatted.index[0] < df_reformatted.index[-1]
    assert df_reformatted.index.freq == "D"
    assert df_reformatted.columns[0] == "value"
    assert df_reformatted.index.tzinfo is not None
    assert len(df_reformatted.columns) == 1


def test_format_energy_data_for_caltrack_billing(il_electricity_cdd_hdd_daily):
    df = il_electricity_cdd_hdd_daily["meter_data"]
    # flipping df
    df = df.reindex(index=df.index[::-1])

    # inserting new value of 0.04 at 09.34 22/11/2015
    new_start = pd.to_datetime("22/11/2015 09:34", dayfirst=True).tz_localize("UTC")
    df.loc[new_start] = [0.04]

    # rename column name to 'consumption'
    df.rename(columns={"value": "consumption"}, inplace=True)

    # df_flipped to pd.Series
    df = df.squeeze()

    df_reformatted = format_energy_data_for_caltrack(df, method="billing")

    assert isinstance(df_reformatted, pd.DataFrame)
    assert df_reformatted.index[0] < df_reformatted.index[-1]
    assert df_reformatted.index.freq == pd.tseries.offsets.MonthEnd()
    assert df_reformatted.columns[0] == "value"
    assert df_reformatted.index.tzinfo is not None
    assert len(df_reformatted.columns) == 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import json

from opendsm.eemeter.samples import load_sample
from opendsm.eemeter.common.transform import get_baseline_data, get_reporting_data
from opendsm.eemeter import (
    DailyBaselineData,
    DailyReportingData,
    DailyModel,
    BillingBaselineData,
    BillingReportingData,
    BillingModel,
    HourlyCaltrackModel,
    HourlyCaltrackBaselineData,
    HourlyCaltrackReportingData,
)


def test_json_daily():
    meter_data, temperature_data, sample_metadata = load_sample(
        "il-electricity-cdd-hdd-daily"
    )

    blackout_start_date = sample_metadata["blackout_start_date"]
    blackout_end_date = sample_metadata["blackout_end_date"]

    # fit baseline model
    baseline_meter_data, warnings = get_baseline_data(
        meter_data, end=blackout_start_date, max_days=365
    )
    baseline_data = DailyBaselineData.from_series(
        baseline_meter_data, temperature_data, is_electricity_data=True
    )
    baseline_model = DailyModel().fit(baseline_data, ignore_disqualification=True)

    # predict on reporting year and calculate savings
    reporting_meter_data, warnings = get_reporting_data(
        meter_data, start=blackout_end_date, max_days=365
    )
    reporting_data = DailyReportingData.from_series(
        reporting_meter_data, temperature_data, is_electricity_data=True
    )
    metered_savings_dataframe = baseline_model.predict(reporting_data)
    total_metered_savings = (
        metered_savings_dataframe["observed"] - metered_savings_dataframe["predicted"]
    ).sum()

    # serialize, deserialize model
    json_str = baseline_model.to_json()
    loaded_model = DailyModel.from_json(json_str)

    # compute metered savings from the loaded model
    prediction_json = loaded_model.predict(reporting_data)
    total_metered_savings_loaded = (
        prediction_json["observed"] - prediction_json["predicted"]
    ).sum()

    # compare results
    assert total_metered_savings == total_metered_savings_loaded


def test_json_billing():
    meter_data, temperature_data, sample_metadata = load_sample(
        "il-electricity-cdd-hdd-billing_monthly"
    )

    blackout_start_date = sample_metadata["blackout_start_date"]
    blackout_end_date = sample_metadata["blackout_end_date"]

    # fit baseline model
    baseline_meter_data, warnings = get_baseline_data(
        meter_data, end=blackout_start_date, max_days=365
    )
    baseline_data = BillingBaselineData.from_series(
        baseline_meter_data, temperature_data, is_electricity_data=True
    )
    baseline_model = BillingModel().fit(baseline_data, ignore_disqualification=True)

    # predict on reporting year and calculate savings
    reporting_meter_data, warnings = get_reporting_data(
        meter_data, start=blackout_end_date, max_days=365
    )
    reporting_data = BillingReportingData.from_series(
        reporting_meter_data, temperature_data, is_electricity_data=True
    )
    metered_savings_dataframe = baseline_model.predict(reporting_data)
    total_metered_savings = (
        metered_savings_dataframe["observed"] - metered_savings_dataframe["predicted"]
    ).sum()

    # serialize, deserialize model
    json_str = baseline_model.to_json()
    loaded_model = BillingModel.from_json(json_str)

    # compute metered savings from the loaded model
    prediction_json = loaded_model.predict(reporting_data)
    total_metered_savings_loaded = (
        prediction_json["observed"] - prediction_json["predicted"]
    ).sum()

    # compare results
    assert total_metered_savings == total_metered_savings_loaded


def test_json_hourly_with_zeros():
    meter_data, temperature_data, sample_metadata = load_sample(
        "il-electricity-cdd-hdd-hourly"
    )
    meter_data["value"] = 0
    baseline = HourlyCaltrackBaselineData.from_series(
        meter_data, temperature_data, is_electricity_data=True
    )
    assert baseline.df["observed"].isnull().all()
    reporting = HourlyCaltrackReportingData.from_series(
        meter_data, temperature_data, is_electricity_data=True
    )
    assert reporting.df["observed"].isnull().all()


def test_json_caltrack_hourly():
    meter_data, temperature_data, sample_metadata = load_sample(
        "il-electricity-cdd-hdd-hourly"
    )

    blackout_start_date = sample_metadata["blackout_start_date"]
    blackout_end_date = sample_metadata["blackout_end_date"]

    # get meter data suitable for fitting a baseline model
    baseline_meter_data, warnings = get_baseline_data(
        meter_data, end=blackout_start_date, max_days=365
    )
    baseline = HourlyCaltrackBaselineData.from_series(
        baseline_meter_data, temperature_data, is_electricity_data=True
    )

    # build a CalTRACK hourly model
    baseline_model = HourlyCaltrackModel().fit(baseline)

    # get a year of reporting period data
    reporting_meter_data, warnings = get_reporting_data(
        meter_data, start=blackout_end_date, max_days=365
    )
    reporting = HourlyCaltrackReportingData.from_series(
        reporting_meter_data, temperature_data, is_electricity_data=True
    )

    result1 = baseline_model.predict(reporting)

    # serialize, deserialize
    json_str = baseline_model.to_json()
    m = HourlyCaltrackModel.from_json(json_str)

    result2 = m.predict(reporting)

    assert result1["predicted"].sum() == result2["predicted"].sum()

    # Check that model metrics are properly serialized/serialized
    assert (
        baseline_model.model.totals_metrics["dec-jan-feb-weighted"].observed_length
        == m.model.totals_metrics["dec-jan-feb-weighted"].observed_length
    )


def test_legacy_deserialization_daily():
    legacy_model_dict = {
        "model_type": "hdd_only",
        "formula": "meter_value ~ hdd_46",
        "status": "QUALIFIED",
        "model_params": {"intercept": 12, "beta_hdd": 2, "heating_balance_point": 50},
        "r_squared_adj": 0.3,
        "warnings": [],
    }
    serialized_str = json.dumps(legacy_model_dict)
    baseline_model = DailyModel.from_2_0_json(serialized_str)

    meter_data, temperature_data, sample_metadata = load_sample(
        "il-electricity-cdd-hdd-daily"
    )
    blackout_end_date = sample_metadata["blackout_end_date"]

    # predict on reporting year and calculate savings
    reporting_meter_data, warnings = get_reporting_data(
        meter_data, start=blackout_end_date, max_days=365
    )
    reporting_data = DailyReportingData.from_series(
        reporting_meter_data, temperature_data, is_electricity_data=True
    )
    metered_savings_dataframe = baseline_model.predict(reporting_data)
    total_metered_savings = (
        metered_savings_dataframe["observed"] - metered_savings_dataframe["predicted"]
    ).sum()

    assert round(total_metered_savings, 2) == 891.2


def test_legacy_deserialization_hourly(request):
    with open(request.fspath.dirname + "/legacy_hourly.json", "r") as f:
        legacy_str = f.read()
    baseline_model = HourlyCaltrackModel.from_2_0_json(legacy_str)

    meter_data, temperature_data, sample_metadata = load_sample(
        "il-electricity-cdd-hdd-hourly"
    )
    blackout_end_date = sample_metadata["blackout_end_date"]

    reporting_meter_data, warnings = get_reporting_data(
        meter_data, start=blackout_end_date, max_days=365
    )
    reporting = HourlyCaltrackReportingData.from_series(
        reporting_meter_data, temperature_data, is_electricity_data=True
    )

    metered_savings_dataframe = baseline_model.predict(reporting)
    total_metered_savings = (
        metered_savings_dataframe["observed"] - metered_savings_dataframe["predicted"]
    ).sum()

    assert round(total_metered_savings, 2) == -52454.02



from __future__ import annotations

import numpy as np

import pydantic

from enum import Enum
from typing import Optional, Literal, Union

import pywt

from opendsm.common.base_settings import BaseSettings



class PCASelection(str, Enum):
    PCA = "pca"
    KERNEL_PCA = "kernel_pca"


class WaveletSelection(str, Enum):
    BIOR1_1 = "bior1.1"
    COIF6 = "coif6"
    COIF17 = "coif17"    # Best error/speed mix
    DB1 = "db1"          # Best error metrics
    DB16 = "db16"
    DB26 = "db26"
    DB29 = "db29"
    HAAR = "haar"
    RBIO1_1 = "rbio1.1"
    SYM11 = "sym11"
    


class WaveletTransformSettings(BaseSettings):
    """wavelet decomposition level"""
    wavelet_n_levels: int = pydantic.Field(
        default=4,
        ge=1,
        # le=5,  #TODO investigate upper limit
    )

    """wavelet choice for wavelet decomposition"""
    wavelet_name: WaveletSelection = pydantic.Field(
        default=WaveletSelection.DB1,
    )

    """signal extension mode for wavelet decomposition"""
    wavelet_mode: str = pydantic.Field(
        default="smooth",
    )

    """PCA method"""
    pca_method: PCASelection = pydantic.Field(
        default=PCASelection.PCA,
    )

    """minimum variance ratio for PCA clustering"""
    pca_min_variance_ratio_explained: Optional[float] = pydantic.Field(
        default=None,
    )

    """number of components to keep for PCA clustering"""
    pca_n_components: Optional[Union[int, Literal["mle"]]] = pydantic.Field(
        default="mle",
    )

    """add mean to pca components"""
    pca_include_median: bool = pydantic.Field(
        default=True,
    )

    @pydantic.model_validator(mode="after")
    def _check_wavelet(self):
        all_wavelets = pywt.wavelist(kind="discrete")
        if self.wavelet_name not in all_wavelets:
            raise ValueError(
                f"'wavelet_name' must be a valid wavelet in PyWavelets: \n{all_wavelets}"
            )

        all_modes = pywt.Modes.modes
        if self.wavelet_mode not in all_modes:
            raise ValueError(
                f"'wavelet_mode' must be a valid mode in PyWavelets: \n{all_modes}"
            )

        return self

    @pydantic.model_validator(mode="after")
    def _check_pca_settings(self):
        if self.pca_n_components is None and self.pca_min_variance_ratio_explained is None:
            raise ValueError(
                "Must specify either 'pca_min_variance_ratio_explained' or 'pca_n_components'"
            )

        if self.pca_n_components is not None:
            if self.pca_min_variance_ratio_explained is not None:
                raise ValueError(
                    "Cannot specify both 'pca_min_variance_ratio_explained' and 'pca_n_components'"
                )
            
            if isinstance(self.pca_n_components, int):
                if self.pca_n_components < 1:
                    raise ValueError(
                        "'pca_n_components' must be >= 1"
                    )

            if (self.pca_n_components == "mle") and (self.pca_method == PCASelection.KERNEL_PCA):
                raise ValueError(
                    "Cannot use 'mle' with 'kernel_pca'"
                )

        if self.pca_min_variance_ratio_explained is not None:
            if not 0.5 <= self.pca_min_variance_ratio_explained <= 1:
                raise ValueError(
                    "'pca_min_variance_ratio_explained' must be between 0.5 and 1"
                )
            
        return self


class ClusterScoringMetric(str, Enum):
    SILHOUETTE = "silhouette"
    SILHOUETTE_MEDIAN = "silhouette_median"
    VARIANCE_RATIO = "variance_ratio"
    DAVIES_BOULDIN = "davies-bouldin"

class DistanceMetric(str, Enum):
    """
    what distance method to use
    """
    EUCLIDEAN = "euclidean"
    SEUCLIDEAN = "seuclidean"
    MANHATTAN = "manhattan"
    COSINE = "cosine"

class ScoreSettings(BaseSettings):
    """minimum cluster size"""
    min_cluster_size: int = pydantic.Field(
        default=1,
        ge=1,
    )

    """scoring method for clustering"""
    score_metric: ClusterScoringMetric = pydantic.Field(
        default=ClusterScoringMetric.VARIANCE_RATIO,
    )

    """distance metric for clustering"""
    distance_metric: DistanceMetric = pydantic.Field(
        default=DistanceMetric.EUCLIDEAN,
    )


class ClusterRangeSettings(BaseSettings):
    """lower bound for number of clusters"""
    lower: int = pydantic.Field(
        default=2,
        ge=2,
    )

    """upper bound for number of clusters"""
    upper: int = pydantic.Field(
        default=24,
        ge=2,
    )

    @pydantic.model_validator(mode="after")
    def _check_n_cluster_range(self):
        if self.lower > self.upper:
            raise ValueError(
                "'n_cluster_lower' must be <= 'n_cluster_upper'"
            )

        return self


class BiKmeansInnerAlgorithms(str, Enum):
    ELKAN = "elkan"
    LLOYD = "lloyd"


class BiKmeansBisectingStrategies(str, Enum):
    BIGGEST_INERTIA = "biggest_inertia"
    LARGEST_CLUSTER = "largest_cluster"


class BisectingKMeansSettings(BaseSettings):
    """number of times to recluster"""
    recluster_count: int = pydantic.Field(
        default=3,
        ge=1,
    )

    """number of times to recluster internally"""
    internal_recluster_count: int = pydantic.Field(
        default=5,
        ge=1,
    )

    """Inner KMeans algorithm used in bisection"""
    inner_algorithm: BiKmeansInnerAlgorithms = pydantic.Field(
        default=BiKmeansInnerAlgorithms.ELKAN,
    )

    """Bisection strategy"""
    bisecting_strategy: BiKmeansBisectingStrategies = pydantic.Field(
        default=BiKmeansBisectingStrategies.LARGEST_CLUSTER,
    )

    n_cluster: ClusterRangeSettings = pydantic.Field(
        default_factory=ClusterRangeSettings
    )

    scoring: ScoreSettings = pydantic.Field(
        default_factory=ScoreSettings
    )

    
class BirchSettings(BaseSettings):
    """radius of the subcluster to merge a new sample in"""
    threshold: float = pydantic.Field(
        default=0.5,
        ge=0,
    )

    """maximum number of CF subclusters in each node"""
    branching_factor: int = pydantic.Field(
        default=50,
        ge=1,
    )

    n_cluster: ClusterRangeSettings = pydantic.Field(
        default_factory=ClusterRangeSettings
    )

    scoring: ScoreSettings = pydantic.Field(
        default_factory=ScoreSettings
    )


class DbscanDistanceAlgorithm(str, Enum):
    AUTO = "auto"
    BRUTE = "brute"
    KD_TREE = "kd_tree"
    BALL_TREE = "ball_tree"

class DBSCANSettings(BaseSettings):
    """maximum distance between two samples for one to be considered as in the neighborhood of the other"""
    epsilon: float = pydantic.Field(
        default=0.5,
        gt=0,
    )

    """minimum number of samples in a neighborhood for a point to be considered as a cluster"""
    min_samples: int = pydantic.Field(
        default=1, # sklearn default is 5
        ge=1,
    )

    """distance metric for calculating distance between samples"""
    distance_metric: DistanceMetric = pydantic.Field(
        default=DistanceMetric.EUCLIDEAN,
    )

    """distance algorithm to use for nearest neighbors"""
    nearest_neighbors_algorithm: DbscanDistanceAlgorithm = pydantic.Field(
        default=DbscanDistanceAlgorithm.AUTO,
    )

    """leaf size for KDTree or BallTree"""
    leaf_size: Optional[int] = pydantic.Field(
        default=30,
    )

    """Minkowski p-norm distance power"""
    minkowski_p: float = pydantic.Field(
        default=2,
        ge=1,
    )


class HdbscanClusterSelectionMethod(str, Enum):
    LEAF = "leaf"
    EXCESS_OF_MASS = "eom"


class HDBSCANSettings(BaseSettings):
    """allow single cluster"""
    allow_single_cluster: bool = pydantic.Field(
        default=True,
    )

    """maximum cluster count"""
    max_cluster_size: Optional[int] = pydantic.Field(
        default=None,
    )

    """minimum number of samples in a group for it to be considered as a cluster"""
    min_samples: int = pydantic.Field(
        default=1,
        ge=1,
    )

    """distance metric for calculating distance between samples"""
    distance_metric: DistanceMetric = pydantic.Field(
        default=DistanceMetric.EUCLIDEAN,
    )

    """samples to calculate distance between neighbors"""
    scoring_sample_count: Optional[int] = pydantic.Field(
        default=None,
    )

    """clusters below this distance threshold will be merged"""
    cluster_selection_epsilon: float = pydantic.Field(
        default=0.0,
        ge=0,
    )

    """distance scaling factor for robust single linkage"""
    robust_single_linkage_scaling: float = pydantic.Field(
        default=1.0,
        gt=0,
    )

    """distance algorithm to use"""
    nearest_neighbors_algorithm: DbscanDistanceAlgorithm = pydantic.Field(
        default=DbscanDistanceAlgorithm.AUTO,
    )

    """leaf size for KDTree or BallTree"""
    leaf_size: Optional[int] = pydantic.Field(
        default=40,
    )

    """cluster selection method"""
    cluster_selection_method: HdbscanClusterSelectionMethod = pydantic.Field(
        default=HdbscanClusterSelectionMethod.EXCESS_OF_MASS,
    )


class SpectralEigenSolver(str, Enum):
    ARPACK = "arpack"
    LOBPCG = "lobpcg"
    # AMG = "amg" # disabled due to additional installation requirements

class AffinityMatrixOptions(str, Enum):
    NEAREST_NEIGHBORS = "nearest_neighbors"
    RBF = "rbf"
    ADDITIVE_CHI2 = "additive_chi2"
    CHI2 = "chi2"
    LINEAR = "linear"
    POLY = "poly"
    POLYNOMIAL = "polynomial"
    LAPLACIAN = "laplacian"
    SIGMOID = "sigmoid"
    COSINE = "cosine"

class SpectralSettings(BaseSettings):
    """eigen solver to use"""
    eigen_solver: Optional[SpectralEigenSolver] = pydantic.Field(
        default=SpectralEigenSolver.ARPACK,
    )

    """number of eigenvectors to use, defaults to n_clusters"""
    n_components: Optional[int] = pydantic.Field(
        default=None,
    )

    """affinity matrix algorithm to use"""
    affinity: AffinityMatrixOptions = pydantic.Field(
        default=AffinityMatrixOptions.RBF,
    )

    """number of nearest neighbors to use for nearest neighbors kernel"""
    nearest_neighbors: int = pydantic.Field(
        default=5,
        ge=1,
    )

    """gamma for RBF, polynomial, sigmoid, laplacian, and chi2 kernels"""
    gamma: float = pydantic.Field(
        default=1.05,
        ge=0, # could be wrong? maybe gt?
    )

    """stopping criterion for eigen decomposition"""
    eigen_tol: Union[float, Literal["auto"]] = pydantic.Field(
        default="auto",
    )

    """label assignment method"""
    assign_labels: Literal["cluster_qr"] = pydantic.Field(
        default="cluster_qr",
    )

    n_cluster: ClusterRangeSettings = pydantic.Field(
        default_factory=ClusterRangeSettings
    )

    scoring: ScoreSettings = pydantic.Field(
        default_factory=ScoreSettings
    )

    @pydantic.model_validator(mode="after")
    def _check_eigen_tol(self):
        if self.eigen_tol != "auto":
            if self.eigen_tol < 0:
                raise ValueError(
                    "'eigen_tol' must be >= 0"
                )

        return self


class SortMethod(str, Enum):
    SIZE = "size"
    PEAK = "peak"
    # VARIANCE = "variance"


class AggregateMethod(str, Enum):
    MEAN = "mean"
    MEDIAN = "median"


class ClusterSortSettings(BaseSettings):
    """sort method"""
    method: SortMethod = pydantic.Field(
        default=SortMethod.PEAK,
    )

    """aggregate method"""
    aggregation: AggregateMethod = pydantic.Field(
        default=AggregateMethod.MEAN
    )

    """sort order"""
    reverse: bool = pydantic.Field(
        default=False,
    )


class ClusterAlgorithms(str, Enum):
    BISECTING_KMEANS = "bisecting_kmeans"
    BIRCH = "birch"
    DBSCAN = "dbscan"
    HDBSCAN = "hdbscan"
    SPECTRAL = "spectral"


class ClusteringSettings(BaseSettings):
    """standardize data boolean"""
    standardize: bool = pydantic.Field(
        default=True,
    )

    """transform settings"""
    transform_settings: Optional[WaveletTransformSettings] = pydantic.Field(
        default_factory=WaveletTransformSettings
    )

    """clustering choice"""
    algorithm_selection: ClusterAlgorithms = pydantic.Field(
        default=ClusterAlgorithms.SPECTRAL,
    )

    """BisectingKMeans settings"""
    bisecting_kmeans: Optional[BisectingKMeansSettings] = pydantic.Field(
        default_factory=BisectingKMeansSettings,
    )

    """Birch settings"""
    birch: Optional[BirchSettings] = pydantic.Field(
        default_factory=BirchSettings,
    )

    """DBSCAN settings"""
    dbscan: Optional[DBSCANSettings] = pydantic.Field(
        default_factory=DBSCANSettings,
    )

    """HDBSCAN settings"""
    hdbscan: Optional[HDBSCANSettings] = pydantic.Field(
        default_factory=HDBSCANSettings,
    )

    """Spectral settings"""
    spectral: Optional[SpectralSettings] = pydantic.Field(
        default_factory=SpectralSettings,
    )

    """sort clusters boolean"""
    sort_clusters: bool = pydantic.Field(
        default=False,
    )

    """sort clusters """
    cluster_sort_options: ClusterSortSettings = pydantic.Field(
        default_factory=ClusterSortSettings,
    )

    """seed for random state assignment"""
    seed: Optional[int] = pydantic.Field(
        default=None,
        ge=0,
    )

    @pydantic.model_validator(mode="after")
    def _check_seed(self):
        if self.seed is None:
            self._seed = np.random.randint(0, 2**32 - 1, dtype=np.int64)
        else:
            self._seed = self.seed

        self._seed = self._seed

        return self

    @pydantic.model_validator(mode="after")
    def _add_standardize_to_transform(self):
        if self.transform_settings is not None:
            self.transform_settings._standardize = self.standardize

        return self

    @pydantic.model_validator(mode="after")
    def _remove_unselected_algorithms(self):
        self.model_config["frozen"] = False

        algo_dict = {
            ClusterAlgorithms.BISECTING_KMEANS: self.bisecting_kmeans,
            ClusterAlgorithms.BIRCH: self.birch,
            ClusterAlgorithms.DBSCAN: self.dbscan,
            ClusterAlgorithms.HDBSCAN: self.hdbscan,
            ClusterAlgorithms.SPECTRAL: self.spectral,
        }

        for k in algo_dict.keys():
            if k != self.algorithm_selection:
                setattr(self, k, None)

        self.model_config["frozen"] = True

        return self


if __name__ == "__main__":
    settings = ClusteringSettings()

    print(settings)

    print(settings._algorithm)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import numpy as np
from opendsm.eemeter.models.daily.base_models.full_model import full_model


def test_full_model_import():
    hdd_bp = 50
    hdd_beta = 0.01
    hdd_k = 0.001
    cdd_bp = 80
    cdd_beta = 0.02
    cdd_k = 0.002
    intercept = 100
    T_fit_bnds = np.array([10, 100]).astype(np.double)
    T = np.linspace(10, 100, 130).astype(np.double)

    res = full_model(
        hdd_bp, hdd_beta, hdd_k, cdd_bp, cdd_beta, cdd_k, intercept, T_fit_bnds, T
    )
    assert res.size == T.size



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from enum import Enum
from typing import Any, Dict, Optional

import numpy as np
from pydantic import BaseModel, ConfigDict


class ModelType(str, Enum):
    # Full model \_/
    HDD_TIDD_CDD_SMOOTH = "hdd_tidd_cdd_smooth"
    HDD_TIDD_CDD = "hdd_tidd_cdd"

    # Heating, temp independent \__
    HDD_TIDD_SMOOTH = "hdd_tidd_smooth"
    HDD_TIDD = "hdd_tidd"

    # Temp independent, cooling __/
    TIDD_CDD_SMOOTH = "tidd_cdd_smooth"
    TIDD_CDD = "tidd_cdd"

    # Temp independent, ___
    TIDD = "tidd"


class ModelCoefficients(BaseModel):
    """
    A class used to represent the coefficients of a model.

    Attributes
    ----------
    model_type : ModelType
        The type of the model.
    intercept : float
        The intercept of the model.
    hdd_bp : float | None
        The heating degree days breakpoint of the model, if applicable.
    hdd_beta : float | None
        The heating degree days beta of the model, if applicable.
    hdd_k : float | None
        The heating degree days k of the model, if applicable.
    cdd_bp : float | None
        The cooling degree days breakpoint of the model, if applicable.
    cdd_beta : float | None
        The cooling degree days beta of the model, if applicable.
    cdd_k : float | None
        The cooling degree days k of the model, if applicable.

    Methods
    -------
    from_np_arrays(coefficients, coefficient_ids)
        Constructs a ModelCoefficients object from numpy arrays of coefficients and their corresponding ids.
    to_np_array()
        Converts the ModelCoefficients object to a numpy array.
    """

    """
    A class used to represent the coefficients of a model.

    Attributes
    ----------
    model_type : ModelType
        The type of the model.
    intercept : float
        The intercept of the model.
    hdd_bp : float | None
        The heating degree days breakpoint of the model, if applicable.
    hdd_beta : float | None
        The heating degree days beta of the model, if applicable.
    hdd_k : float | None
        The heating degree days k of the model, if applicable.
    cdd_bp : float | None
        The cooling degree days breakpoint of the model, if applicable.
    cdd_beta : float | None
        The cooling degree days beta of the model, if applicable.
    cdd_k : float | None
        The cooling degree days k of the model, if applicable.

    Methods
    -------
    from_np_arrays(coefficients, coefficient_ids)
        Constructs a ModelCoefficients object from numpy arrays of coefficients and their corresponding ids.
    to_np_array()
        Converts the ModelCoefficients object to a numpy array.
    """

    model_type: ModelType
    intercept: float
    hdd_bp: Optional[float] = None
    hdd_beta: Optional[float] = None
    hdd_k: Optional[float] = None
    cdd_bp: Optional[float] = None
    cdd_beta: Optional[float] = None
    cdd_k: Optional[float] = None

    # suppress namespace warning for model_type
    model_config = ConfigDict(protected_namespaces=())

    @property
    def is_smooth(self):
        return self.model_type in [
            ModelType.HDD_TIDD_CDD_SMOOTH,
            ModelType.HDD_TIDD_SMOOTH,
            ModelType.TIDD_CDD_SMOOTH,
        ]

    @property
    def model_key(self):
        """Used inside OptimizedResult when reducing model"""
        if self.model_type == ModelType.HDD_TIDD_CDD_SMOOTH:
            return "hdd_tidd_cdd_smooth"
        elif self.model_type == ModelType.HDD_TIDD_CDD:
            return "hdd_tidd_cdd"
        elif self.model_type in [ModelType.HDD_TIDD_SMOOTH, ModelType.TIDD_CDD_SMOOTH]:
            return "c_hdd_tidd_smooth"
        elif self.model_type in [ModelType.HDD_TIDD, ModelType.TIDD_CDD]:
            return "c_hdd_tidd"
        elif self.model_type == ModelType.TIDD:
            return "tidd"

    @classmethod
    def from_np_arrays(cls, coefficients, coefficient_ids):
        """
        This class method creates a ModelCoefficients instance from numpy arrays of coefficients and their corresponding ids.

        Args:
            cls (class): The class to which this class method belongs.
            coefficients (np.array): A numpy array of coefficients.
            coefficient_ids (list): A list of coefficient ids.

        Returns:
            ModelCoefficients: An instance of ModelCoefficients class.

        Raises:
            ValueError: If the coefficient_ids do not match any of the expected patterns.

        The method matches the coefficient_ids to predefined patterns and based on the match,
        it initializes a ModelCoefficients instance with the corresponding model_type and coefficients.
        If the coefficient_ids do not match any of the predefined patterns, it raises a ValueError.
        """

        if coefficient_ids == [
            "hdd_bp",
            "hdd_beta",
            "hdd_k",
            "cdd_bp",
            "cdd_beta",
            "cdd_k",
            "intercept",
        ]:
            hdd_bp = coefficients[0]
            hdd_beta = coefficients[1]
            hdd_k = coefficients[2]
            cdd_bp = coefficients[3]
            cdd_beta = coefficients[4]
            cdd_k = coefficients[5]
            if cdd_bp < hdd_bp:
                hdd_bp, cdd_bp = cdd_bp, hdd_bp
                hdd_beta, cdd_beta = cdd_beta, hdd_beta
                hdd_k, cdd_k = cdd_k, hdd_k
            return ModelCoefficients(
                model_type=ModelType.HDD_TIDD_CDD_SMOOTH,
                hdd_bp=hdd_bp,
                hdd_beta=hdd_beta,
                hdd_k=hdd_k,
                cdd_bp=cdd_bp,
                cdd_beta=cdd_beta,
                cdd_k=cdd_k,
                intercept=coefficients[6],
            )
        elif coefficient_ids == [
            "hdd_bp",
            "hdd_beta",
            "cdd_bp",
            "cdd_beta",
            "intercept",
        ]:
            hdd_bp = coefficients[0]
            hdd_beta = coefficients[1]
            cdd_bp = coefficients[2]
            cdd_beta = coefficients[3]
            if cdd_bp < hdd_bp:
                hdd_bp, cdd_bp = cdd_bp, hdd_bp
                hdd_beta, cdd_beta = cdd_beta, hdd_beta
            return ModelCoefficients(
                model_type=ModelType.HDD_TIDD_CDD,
                hdd_bp=hdd_bp,
                hdd_beta=hdd_beta,
                cdd_bp=cdd_bp,
                cdd_beta=cdd_beta,
                intercept=coefficients[4],
            )
        elif coefficient_ids == [
            "c_hdd_bp",
            "c_hdd_beta",
            "c_hdd_k",
            "intercept",
        ]:
            if coefficients[1] < 0:  # model is heating dependent
                hdd_bp = coefficients[0]
                hdd_beta = coefficients[1]
                hdd_k = coefficients[2]
                cdd_bp = cdd_beta = cdd_k = None
                model_type = ModelType.HDD_TIDD_SMOOTH
            else:  # model is cooling dependent
                cdd_bp = coefficients[0]
                cdd_beta = coefficients[1]
                cdd_k = coefficients[2]
                hdd_bp = hdd_beta = hdd_k = None
                model_type = ModelType.TIDD_CDD_SMOOTH
            return ModelCoefficients(
                model_type=model_type,
                hdd_bp=hdd_bp,
                hdd_beta=hdd_beta,
                hdd_k=hdd_k,
                cdd_bp=cdd_bp,
                cdd_beta=cdd_beta,
                cdd_k=cdd_k,
                intercept=coefficients[3],
            )
        elif coefficient_ids == [
            "c_hdd_bp",
            "c_hdd_beta",
            "intercept",
        ]:
            if coefficients[1] < 0:  # model is heating dependent
                hdd_bp = coefficients[0]
                hdd_beta = coefficients[1]
                cdd_bp = cdd_beta = None
                model_type = ModelType.HDD_TIDD
            else:  # model is cooling dependent
                cdd_bp = coefficients[0]
                cdd_beta = coefficients[1]
                hdd_bp = hdd_beta = None
                model_type = ModelType.TIDD_CDD
            return ModelCoefficients(
                model_type=model_type,
                hdd_bp=hdd_bp,
                hdd_beta=hdd_beta,
                cdd_bp=cdd_bp,
                cdd_beta=cdd_beta,
                intercept=coefficients[2],
            )
        elif coefficient_ids == [
            "intercept",
        ]:
            return ModelCoefficients(
                model_type=ModelType.TIDD,
                intercept=coefficients[0],
            )
        else:
            raise ValueError

    def to_np_array(self):
        """
        This method converts the model parameters into a numpy array based on the model type.

        The model type determines which parameters are included in the array. The parameters are:
        - hdd_bp: The base point for heating degree days (HDD)
        - hdd_beta: The beta coefficient for HDD
        - hdd_k: The smoothing parameter for HDD
        - cdd_bp: The base point for cooling degree days (CDD)
        - cdd_beta: The beta coefficient for CDD
        - cdd_k: The smoothing parameter for CDD
        - intercept: The model's intercept

        Returns:
            np.array: A numpy array containing the relevant parameters for the model type.
        """

        if self.model_type == ModelType.HDD_TIDD_CDD_SMOOTH:
            return np.array(
                [
                    self.hdd_bp,
                    self.hdd_beta,
                    self.hdd_k,
                    self.cdd_bp,
                    self.cdd_beta,
                    self.cdd_k,
                    self.intercept,
                ]
            )
        elif self.model_type == ModelType.HDD_TIDD_CDD:
            return np.array(
                [
                    self.hdd_bp,
                    self.hdd_beta,
                    self.cdd_bp,
                    self.cdd_beta,
                    self.intercept,
                ]
            )
        elif self.model_type == ModelType.HDD_TIDD_SMOOTH:
            return np.array([self.hdd_bp, self.hdd_beta, self.hdd_k, self.intercept])
        elif self.model_type == ModelType.TIDD_CDD_SMOOTH:
            return np.array([self.cdd_bp, self.cdd_beta, self.cdd_k, self.intercept])
        elif self.model_type == ModelType.HDD_TIDD:
            return np.array([self.hdd_bp, self.hdd_beta, self.intercept])
        elif self.model_type == ModelType.TIDD_CDD:
            return np.array([self.cdd_bp, self.cdd_beta, self.intercept])
        elif self.model_type == ModelType.TIDD:
            return np.array([self.intercept])


class DailySubmodelParameters(BaseModel):
    coefficients: ModelCoefficients
    temperature_constraints: Dict[str, float]
    f_unc: float

    @property
    def model_type(self):
        return self.coefficients.model_type


class DailyModelParameters(BaseModel):
    submodels: Dict[str, DailySubmodelParameters]
    info: Optional[Dict[str, Any]]
    settings: Optional[Dict[str, Any]]

    @classmethod
    def from_2_0_params(cls, data):
        model_2_0 = data.get("model_type")
        if model_2_0 == "intercept_only":
            model_type = ModelType.TIDD
        elif model_2_0 == "hdd_only":
            model_type = ModelType.HDD_TIDD
        elif model_2_0 == "cdd_only":
            model_type = ModelType.TIDD_CDD
        elif model_2_0 == "cdd_hdd":
            model_type = ModelType.HDD_TIDD_CDD
        elif model_2_0 is None:
            raise ValueError("Missing model type")
        else:
            raise ValueError(f"Unknown model type: {model_2_0}")
        params = data["model_params"]

        hdd_beta = params.get("beta_hdd")
        if model_type == ModelType.HDD_TIDD:
            # sign is reversed for heating-only model
            hdd_beta *= -1

        daily_coeffs = ModelCoefficients(
            model_type=model_type,
            intercept=params.get("intercept"),
            hdd_bp=params.get("heating_balance_point"),
            hdd_beta=hdd_beta,
            cdd_bp=params.get("cooling_balance_point"),
            cdd_beta=params.get("beta_cdd"),
        )
        submodel_params = DailySubmodelParameters(
            coefficients=daily_coeffs,
            temperature_constraints={
                "T_min": -100,
                "T_min_seg": -100,
                "T_max": 200,
                "T_max_seg": 200,
            },
            f_unc=np.inf,
        )
        return cls(
            # TODO handle settings correctly with something in config.py
            settings={"from 2.0 - will fail if attempting from_dict()": True},
            info={},
            submodels={
                # no splits, full calendar
                "fw-su_sh_wi": submodel_params,
            },
        )



from __future__ import annotations

import numpy as np
import pandas as pd

import pydantic

from enum import Enum
from typing import Optional, Literal, Union, TypeVar, Dict

import pywt

from opendsm.common.base_settings import BaseSettings
from opendsm.common.clustering.settings import ClusteringSettings
from opendsm.common.metrics import BaselineMetrics

from opendsm.eemeter.common.warnings import EEMeterWarning

# from opendsm.common.const import CountryCode


class SelectionChoice(str, Enum):
    CYCLIC = "cyclic"
    RANDOM = "random"


class ScalingChoice(str, Enum):
    ROBUST_SCALER = "robustscaler"
    STANDARD_SCALER = "standardscaler"


class BinningChoice(str, Enum):
    EQUAL_SAMPLE_COUNT = "equal_sample_count"
    EQUAL_BIN_WIDTH = "equal_bin_width"
    SET_BIN_WIDTH = "set_bin_width"
    FIXED_BINS = "fixed_bins"


class DefaultTrainingFeatures(str, Enum):
    SOLAR = ["temperature", "ghi"]
    NONSOLAR = ["temperature"]


class AggregationMethod(str, Enum):
    MEAN = "mean"
    MEDIAN = "median"


class BaseModel(str, Enum):
    ELASTICNET = "elasticnet"
    KERNEL_RIDGE = "kernel_ridge"


class TemperatureBinSettings(BaseSettings):
    """how to bin temperature data"""
    method: BinningChoice = pydantic.Field(
        default=BinningChoice.FIXED_BINS,
    )

    """number of temperature bins"""
    n_bins: Optional[int] = pydantic.Field(
        default=None,
        ge=1,
    )

    """temperature bin width in fahrenheit"""
    bin_width: Optional[float] = pydantic.Field(
        default=25,
        ge=1,
    )

    """specified fixed temperature bins in fahrenheit"""
    fixed_bins: Optional[list[float]] = pydantic.Field(
        default=[10, 30, 50, 65, 75, 90, 105],
    )

    "minimum bin count"
    min_bin_count: Optional[int] = pydantic.Field(
        default=20,
        ge=1,
    )

    """use edge bins bool"""
    include_edge_bins: bool = pydantic.Field(
        default=True, 
    )

    """rate for edge temperature bins"""
    edge_bin_rate: Optional[Union[float, Literal["heuristic"]]] = pydantic.Field(
        default="heuristic",
    )

    """percent of total data in edge bins"""
    edge_bin_percent: Optional[float] = pydantic.Field(
        default=None,
        gt=0,
        le=0.45,
    )

    """offset normalized temperature range for edge bins (keeps exp from blowing up)"""
    edge_bin_temperature_range_offset: Optional[float] = pydantic.Field(
        default=1.0, # prior 1.0
        ge=0,
    )

    @pydantic.model_validator(mode="after")
    def _check_temperature_bins(self):
        if self.method == BinningChoice.EQUAL_SAMPLE_COUNT:
            if self.n_bins is None:
                raise ValueError(
                    "'n_bins' must be specified if 'method' is 'equal_sample_count'."
                )
            if self.n_bins < 1:
                raise ValueError("'n_bins' must be greater than 0.")

        elif self.method == BinningChoice.EQUAL_BIN_WIDTH:
            if self.bin_width is None:
                raise ValueError(
                    "'bin_width' must be specified if 'method' is 'equal_bin_width'."
                )
            if self.bin_width < 1:
                raise ValueError("'bin_width' must be greater than 0.")

        elif self.method == BinningChoice.SET_BIN_WIDTH:
            if self.bin_width is None:
                raise ValueError(
                    "'bin_width' must be specified if 'method' is 'set_bin_width'."
                )
            if self.bin_width < 1:
                raise ValueError("'bin_width' must be greater than 0.")

        elif self.method == BinningChoice.FIXED_BINS:
            if self.fixed_bins is None:
                raise ValueError(
                    "'fixed_bins' must be specified if 'method' is 'fixed_bins'."
                )

        else:
            raise ValueError(f"Invalid method: {self.method}")

        return self

    @pydantic.model_validator(mode="after")
    def _check_edge_bins(self):
        if self.include_edge_bins:
            if self.edge_bin_rate is None:
                raise ValueError(
                    "'edge_bin_rate' must be specified if 'include_edge_bins' is True."
                )
            if self.edge_bin_percent is None and self.method != BinningChoice.FIXED_BINS:
                raise ValueError(
                    "'edge_bin_days' must be specified if 'include_edge_bins' is True."
                )
            if self.edge_bin_temperature_range_offset is None:
                raise ValueError(
                    "'edge_bin_temperature_range_offset' must be specified if 'include_edge_bins' is True."
                )

        else:
            if self.edge_bin_rate is not None:
                raise ValueError(
                    "'edge_bin_rate' must be None if 'include_edge_bins' is False."
                )
            if self.edge_bin_percent is not None:
                raise ValueError(
                    "'edge_bin_days' must be None if 'include_edge_bins' is False."
                )
            if self.edge_bin_temperature_range_offset is not None:
                raise ValueError(
                    "'edge_bin_temperature_range_offset' must be None if 'include_edge_bins' is False."
                )

        return self


class ElasticNetSettings(BaseSettings):
    """ElasticNet alpha parameter"""

    alpha: float = pydantic.Field(
        default=0.0139,
        ge=0,
    )

    """ElasticNet l1_ratio parameter"""
    l1_ratio: float = pydantic.Field(
        default=0.871,
        ge=0,
        le=1,
    )

    """ElasticNet fit_intercept parameter"""
    fit_intercept: bool = pydantic.Field(
        default=True,
    )

    """ElasticNet parameter to precompute Gram matrix"""
    precompute: bool = pydantic.Field(
        default=False,
    )

    """ElasticNet max_iter parameter"""
    max_iter: int = pydantic.Field(
        default=3000,
        ge=1,
        le=2**32 - 1,
    )

    """ElasticNet copy_X parameter"""
    copy_x: bool = pydantic.Field(
        default=True,
    )

    """ElasticNet tol parameter"""
    tol: float = pydantic.Field(
        default=1e-3,
        gt=0,
    )

    """ElasticNet selection parameter"""
    selection: SelectionChoice = pydantic.Field(
        default=SelectionChoice.CYCLIC,
    )

    """ElasticNet warm_start parameter"""
    warm_start: bool = pydantic.Field(
        default=False,
    )


class KernelRidgeSettings(BaseSettings):
    """Kernel Ridge alpha parameter"""
    alpha: float = pydantic.Field(
        default=0.0425,
        ge=0,
    )

    """Kernel Ridge kernel parameter"""
    kernel: str = pydantic.Field(
        default="rbf",
    )

    """Kernel Ridge gamma parameter"""
    gamma: Optional[float] = pydantic.Field(
        default=None,
        gt=0,
    )


class CAlgoChoice(str, Enum):
    IQR_LEGACY = "iqr_legacy"
    IQR = "iqr"
    MAD = "mad"
    STDEV = "stdev"


class AdaptiveWeightsSettings(BaseSettings):
    """Adaptive Weights for ElasticNet"""
    enabled: bool = pydantic.Field(
        default=True,
    )

    """Sigma threshold for calculating C"""
    sigma: Optional[float] = pydantic.Field(
        default=4.55,
        gt=0,
    )

    """Adaptive weights window size"""
    window_size: Optional[int] = pydantic.Field(
        default=3,
        ge=1,
        le=12,
    )

    """Algorithm to use for calculating C"""
    c_algo: Optional[CAlgoChoice] = pydantic.Field(
        default=CAlgoChoice.IQR,
    )

    """Number of iterations to iterate weights"""
    max_iter: Optional[int] = pydantic.Field(
        default=100,   # Exits early based on tol
        ge=1,
    )

    """Relative difference in weights to stop iteration"""
    tol: Optional[float] = pydantic.Field(
        default=1E-3,   # Previously was using 1e-4
        ge=0,
    )

    @pydantic.model_validator(mode="after")
    def _check_adaptive_weights(self):
        if self.enabled:
            # iterate through all the parameters to check if they are set
            # if any are None, raise an error
            pass
        else:
            # iterate through all the parameters to check if they are set
            # if any are not None, raise an error
            pass

        return self


class Criterion(str, Enum):
    AIC = "aic"
    BIC = "bic"


# analytic_features = ['GHI', 'Temperature', 'DHI', 'DNI', 'Relative Humidity', 'Wind Speed', 'Clearsky DHI', 'Clearsky DNI', 'Clearsky GHI', 'Cloud Type']
class BaseHourlySettings(BaseSettings):
    """train features used within the model"""

    train_features: Optional[list[str]] = None

    """CVRMSE threshold for model disqualification"""
    cvrmse_threshold: float = pydantic.Field(
        default=1.4,
    )

    """PNRMSE threshold for model disqualification"""
    pnrmse_threshold: float = pydantic.Field(
        default=2.2,
    )

    """minimum number of training hours per day below which a day is excluded"""
    min_daily_training_hours: int = pydantic.Field(
        default=12,
        ge=0,
        le=24,
    )

    """temperature bin settings"""
    temperature_bin: Optional[TemperatureBinSettings] = pydantic.Field(
        default_factory=TemperatureBinSettings,
    )

    """settings for temporal clustering"""
    temporal_cluster: ClusteringSettings = pydantic.Field(
        default_factory=ClusteringSettings,
    )

    """temporal cluster aggregation method"""
    temporal_cluster_aggregation: AggregationMethod = pydantic.Field(
        default=AggregationMethod.MEDIAN,
    )

    """temporal cluster/temperature bin/temperature interaction scalar"""
    interaction_scalar: float = pydantic.Field(
        default=0.524,
        gt=0,
    )

    """supplemental time series column names"""
    supplemental_time_series_columns: Optional[list] = pydantic.Field(
        default=None,
    )

    """supplemental categorical column names"""
    supplemental_categorical_columns: Optional[list] = pydantic.Field(
        default=None,
    )

    """base model type"""
    base_model: BaseModel = pydantic.Field(
        default=BaseModel.ELASTICNET,
    )

    """ElasticNet settings"""
    elasticnet: Optional[ElasticNetSettings] = pydantic.Field(
        default_factory=ElasticNetSettings,
    )

    """Kernel Ridge settings"""
    kernel_ridge: Optional[KernelRidgeSettings] = pydantic.Field(
        default_factory=KernelRidgeSettings,
    )

    """Adaptive Weights settings"""
    adaptive_weights: AdaptiveWeightsSettings = pydantic.Field(
        default_factory=AdaptiveWeightsSettings,
    )

    """Feature scaling method"""
    scaling_method: ScalingChoice = pydantic.Field(
        default=ScalingChoice.STANDARD_SCALER,
    )

    """seed for any random state assignment (ElasticNet, Clustering)"""
    seed: Optional[int] = pydantic.Field(
        default=None,
        ge=0,
    )

    @pydantic.model_validator(mode="after")
    def _check_seed(self):
        if self.seed is None:
            self._seed = np.random.randint(0, 2**32 - 1, dtype=np.int64)
        else:
            self._seed = self.seed

        self.elasticnet._seed = self._seed
        self.temporal_cluster._seed = self._seed

        return self

    @pydantic.model_validator(mode="after")
    def _remove_unselected_model_settings(self):
        self.model_config["frozen"] = False
        
        if self.base_model == BaseModel.ELASTICNET:
            self.kernel_ridge = None
        elif self.base_model == BaseModel.KERNEL_RIDGE:
            self.elasticnet = None

        self.model_config["frozen"] = True

        return self

    def add_default_features(self, incoming_columns: list[str]):
        """ "called prior fit step to set default training features"""
        if "ghi" in incoming_columns:
            default_features = ["temperature", "ghi"]
        else:
            default_features = ["temperature"]
        return self.model_copy(update={"train_features": default_features})


class HourlySolarSettings(BaseHourlySettings):
    """train features used within the model"""

    train_features: list[str] = pydantic.Field(
        default=["temperature", "ghi"],
    )

    @pydantic.field_validator("train_features", mode="after")
    def _add_required_features(cls, v):
        required_features = ["ghi", "temperature"]
        for feature in required_features:
            if feature not in v:
                v.insert(0, feature)
        return v


class HourlyNonSolarSettings(BaseHourlySettings):
    """number of temperature bins"""

    # TEMPERATURE_BIN_COUNT: Optional[int] = pydantic.Field(
    #     default=10,
    #     ge=1,
    # )
    train_features: list[str] = pydantic.Field(
        default=["temperature"],
    )

    @pydantic.field_validator("train_features", mode="after")
    def _add_required_features(cls, v):
        if "temperature" not in v:
            v.insert(0, "temperature")
        return v


class ModelInfo(pydantic.BaseModel):
    """additional information about the model"""

    warnings: list[EEMeterWarning]
    disqualification: list[EEMeterWarning]
    error: dict
    baseline_timezone: str
    version: str


class SerializeModel(BaseSettings):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    settings: Optional[BaseHourlySettings] = None
    temporal_clusters: Optional[list[list[int]]] = None
    temperature_bin_edges: Optional[list] = None
    temperature_edge_bin_coefficients: Optional[Dict[int, Dict[str, float]]] = None
    ts_features: Optional[list] = None
    categorical_features: Optional[list] = None
    feature_scaler: Optional[Dict[str, list[float]]] = None
    catagorical_scaler: Optional[Dict[str, list[float]]] = None
    y_scaler: Optional[list[float]] = None
    coefficients: Optional[list[list[float]]] = None
    intercept: Optional[list[float]] = None
    baseline_metrics: Optional[BaselineMetrics] = None
    info: ModelInfo



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import numba
import numpy as np
from scipy.optimize import minimize_scalar
from scipy.special import lambertw
from scipy.stats import linregress, theilslopes

from opendsm.common.utils import OoM_numba


def get_intercept(y, alpha=2):
    """
    Calculates the intercept of a linear regression model.

    Parameters:
    -----------
    y : array-like
        Dependent variable.
    alpha : float, optional
        Significance level for the Theil-Sen estimator. Default is 2.

    Returns:
    --------
    intercept : float
        Intercept of the linear regression model.
    """

    if alpha == 2:
        intercept = np.mean(y)
    else:
        intercept = np.median(y)

    return intercept


def get_slope(x, y, x_bp, intercept, alpha=2):
    def slope_fcn_dec(x, y, x_bp, intercept, alpha):
        def slope_fcn(slope):  # TODO: This function could be numba'd
            model = slope * (x - x_bp) + intercept
            resid = y - model

            if alpha == 2:
                obj = np.sqrt(np.sum(resid**2))
            else:
                # obj = np.sum(np.abs(resid)) # MAE
                obj = np.sum(np.logaddexp(resid, -resid) - np.log(2))  # log cosh

            return obj

        return slope_fcn

    opt_fcn = slope_fcn_dec(x, y, x_bp, intercept, alpha)

    slope = minimize_scalar(opt_fcn, method="brent", tol=0.1).x

    return slope


def linear_fit(x, y, alpha):
    if len(set(x)) == 1:
        slope = 0
        intercept = x[0]
    elif alpha == 2:
        res = linregress(x, y)

        slope = res.slope
        intercept = res.intercept
    else:
        slope, intercept, _, _ = theilslopes(y, x, alpha=0.95)

    return slope, intercept


# smoothed curve will match unsmoothed at the perc_match% decay of the exp
# if pct_match == 1.0, then it will converge at - or + inf
def get_smooth_coeffs(hdd_bp, pct_hdd_k, cdd_bp, pct_cdd_k, min_pct_k=0.01):
    if (pct_hdd_k < min_pct_k) and (pct_cdd_k < min_pct_k):
        return np.array([hdd_bp, 0, cdd_bp, 0])

    pct_match = 1
    hdd_w = cdd_w = 0
    if pct_match < 1:
        hdd_w = lambertw((1 - pct_match) / np.exp(1)).real
        # cdd_w = lambertw(-(1 - pct_match)/np.exp(1)).real
        cdd_w = -hdd_w

    pct_k_sum = pct_hdd_k + pct_cdd_k
    if pct_k_sum > 1:
        pct_hdd_k /= pct_k_sum
        pct_cdd_k /= pct_k_sum

    # calculate the smoothing parameter as a percentage of the maximum allowed k
    hdd_k = pct_hdd_k * (cdd_bp - hdd_bp) / (1 - hdd_w)
    cdd_k = pct_cdd_k * (cdd_bp - hdd_bp) / (1 + cdd_w)

    # move breakpoints based on k
    hdd_bp = hdd_bp + hdd_k * (1 - hdd_w)
    cdd_bp = cdd_bp - cdd_k * (1 + cdd_w)

    return np.array([hdd_bp, hdd_k, cdd_bp, cdd_k])


@numba.jit(nopython=True, error_model="numpy", cache=True)
def fix_identical_bnds(bnds):
    for i in np.argwhere(bnds[:, 0] == bnds[:, 1]):
        bnds[i, :] = bnds[i, :] + np.array([-1.0, 1.0]) * 10 ** OoM_numba(
            bnds[i, 0], method="floor"
        )

    return bnds


def get_T_bnds(T, settings):
    n_min_seg = settings.segment_minimum_count

    T_min = np.min(T)
    T_max = np.max(T)
    T_min_seg = np.partition(T, n_min_seg)[n_min_seg]
    T_max_seg = np.partition(T, -n_min_seg)[-n_min_seg]

    return [T_min, T_max], [T_min_seg, T_max_seg]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from pathlib import Path
import copy
from typing import Optional, Union
from datetime import date

from opendsm.eemeter.common.data_processor_utilities import (
    compute_minimum_granularity,
    remove_duplicates,
)
from opendsm.common.hourly_interpolation import interpolate
from opendsm.eemeter.common.features import compute_temperature_features
from opendsm.eemeter.common.sufficiency_criteria import HourlySufficiencyCriteria
from opendsm.eemeter.common.warnings import EEMeterWarning

import numpy as np
import pandas as pd


class NREL_Weather_API:
    api_key = (
        "---"  # get your own key from https://developer.nrel.gov/signup/  #Required
    )
    name = "---"  # required
    email = "---"  # required
    interval = "60"  # required

    attributes = "ghi,dhi,dni,wind_speed,air_temperature,cloud_type,dew_point,clearsky_dhi,clearsky_dni,clearsky_ghi"  # not required
    leap_year = "false"  # not required
    utc = "false"  # not required
    reason_for_use = "---"  # not required
    your_affiliation = "---"  # not required
    mailing_list = "false"  # not required

    # cache = Path("/app/.recurve_cache/data/MCE/MCE_weather_stations")
    cache = Path("/app/.recurve_cache/data/MCE/Weather_stations")

    use_cache = True

    round_minutes_method = "floor"  # [None, floor, ceil, round]

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
        self.cache.mkdir(parents=True, exist_ok=True)

    def get_data(self, lat, lon, years=[2017, 2021]):
        data_path = self.cache / f"{lat}_{lon}.pkl"
        if data_path.exists() and self.use_cache:
            df = pd.read_pickle(data_path)

        else:
            years = list(range(min(years), max(years) + 1))

            df = self.query_API(lat, lon, years)

            df.columns = [x.lower().replace(" ", "_") for x in df.columns]

            if self.round_minutes_method == "floor":
                df["datetime"] = df["datetime"].dt.floor("h")
            elif self.round_minutes_method == "ceil":
                df["datetime"] = df["datetime"].dt.ceil("h")
            elif self.round_minutes_method == "round":
                df["datetime"] = df["datetime"].dt.round("h")

            df = df.set_index("datetime")

            if self.use_cache:
                df.to_pickle(data_path)

        return df

    def query_API(self, lat, lon, years):
        leap_year = self.leap_year
        interval = self.interval
        utc = self.utc
        api_key = self.api_key
        name = self.name
        email = self.email

        year_df = []
        for year in years:
            year = str(year)

            url = self._generate_url(
                lat, lon, year, leap_year, interval, utc, api_key, name, email
            )
            df = pd.read_csv(url, skiprows=2)

            # Set the time index in the pandas dataframe:
            # set datetime using the year, month, day, and hour
            df["datetime"] = pd.to_datetime(
                df[["Year", "Month", "Day", "Hour", "Minute"]]
            )

            df = df.drop(columns=["Year", "Month", "Day", "Hour", "Minute"])
            df = df.dropna()

            year_df.append(df)

        # merge the dataframes for different years
        df = pd.concat(year_df, axis=0)

        return df

    def _generate_url(
        self, lat, lon, year, leap_year, interval, utc, api_key, name, email
    ):
        query = f"?wkt=POINT({lon}%20{lat})&names={year}&interval={interval}&api_key={api_key}&full_name={name}&email={email}&utc={utc}"

        if year == "2021":
            # details: https://developer.nrel.gov/docs/solar/nsrdb/psm3-2-2-download/
            url = f"https://developer.nrel.gov/api/nsrdb/v2/solar/psm3-2-2-download.csv{query}"

        elif year in [str(i) for i in range(1998, 2021)]:
            # details: https://developer.nrel.gov/docs/solar/nsrdb/psm3-download/
            url = f"https://developer.nrel.gov/api/nsrdb/v2/solar/psm3-download.csv{query}"

        else:
            print("Year must be between 1998 and 2021")
            url = None

        return url


class _HourlyData:
    """Private base class for hourly baseline and reporting data

    Will raise exception during data sufficiency check if instantiated
    """

    def __init__(
        self,
        df: pd.DataFrame,
        is_electricity_data: bool,
        pv_start: Union[date, str, None] = None,
        **kwargs: dict,
    ):
        self._df = None
        self.warnings = []
        self.disqualification = []
        self.is_electricity_data = is_electricity_data
        self.tz = None

        # TODO copied from HourlyData
        self._to_be_interpolated_columns = []
        self._outputs = []

        self.pv_start = None
        if pv_start is not None:
            self.pv_start = pd.to_datetime(pv_start).date()

        # TODO not sure why we're keeping this copy, just set the attrs
        self._kwargs = copy.deepcopy(kwargs)
        if "outputs" in self._kwargs:
            self._outputs = copy.deepcopy(self._kwargs["outputs"])
        else:
            self._outputs = ["temperature", "observed"]

        self._df = self._set_data(df)
        disqualification, warnings = self._check_data_sufficiency()

        self.disqualification += disqualification
        self.warnings += warnings
        self.log_warnings()

    @property
    def df(self):
        """Get the corrected input data stored in the class. The actual dataframe is immutable, this returns a copy."""
        if self._df is None:
            return None
        else:
            return self._df.copy()

    def log_warnings(self):
        """
        Logs the warnings and disqualifications associated with the data.
        """
        for warning in self.warnings + self.disqualification:
            warning.warn()

    def _get_contiguous_datetime(self, df):
        # get earliest datetime and latest datetime
        # make earliest start at 0 and latest end at 23, this ensures full days
        earliest_datetime = df.index.min().replace(
            hour=0, minute=0, second=0, microsecond=0
        )
        latest_datetime = df.index.max().replace(
            hour=23, minute=0, second=0, microsecond=0
        )

        # create a new index with all the hours between the earliest and latest datetime
        complete_dt = pd.date_range(
            start=earliest_datetime, end=latest_datetime, freq="h"
        )

        # merge meter data with complete_dt
        df = df.reindex(complete_dt)

        df["date"] = df.index.date
        df["hour_of_day"] = df.index.hour

        return df

    def _interpolate(self, df):
        # make column of interpolated boolean if any observed or temperature is nan
        # check if in each row of the columns in output has nan values, the interpolated column will be true
        if "to_be_interpolated_columns" in self._kwargs:
            self._to_be_interpolated_columns = self._kwargs[
                "to_be_interpolated_columns"
            ].copy()
            self._outputs += [
                f"{col}"
                for col in self._to_be_interpolated_columns
                if col not in self._outputs
            ]
        else:
            self._to_be_interpolated_columns = ["temperature", "observed"]
            if "ghi" in df.columns:
                self._to_be_interpolated_columns.append("ghi")

        for col in self._to_be_interpolated_columns:
            if f"interpolated_{col}" in df.columns:
                continue
            self._outputs += [f"interpolated_{col}"]

        # we can add kwargs to the interpolation class like: inter_kwargs = {"n_cor_idx": self.kwargs["n_cor_idx"]}
        df = interpolate(df, columns=self._to_be_interpolated_columns)

        return df

    def _add_pv_start_date(self, df, model_type="TS"):
        if self.pv_start is None:
            self.pv_start = df.index.date.min()

        if "ts" in model_type.lower() or "time" in model_type.lower():
            df["has_pv"] = 0
            df.loc[df["date"] >= self.pv_start, "has_pv"] = 1

        else:
            df["has_pv"] = False
            df.loc[df["date"] >= self.pv_start, "has_pv"] = True
        return df

    def _merge_meter_temp(self, meter, temp):
        df = meter.merge(
            temp, left_index=True, right_index=True, how="left"
        ).tz_convert(meter.index.tz)
        return df

    def _check_data_sufficiency(self):
        raise NotImplementedError(
            "Can't instantiate class _HourlyData, use HourlyBaselineData or HourlyReportingData."
        )

    def _set_data(self, data: pd.DataFrame):
        df = data.copy()
        expected_columns = [
            "observed",
            "temperature",
            # "ghi",
        ]
        if not set(expected_columns).issubset(set(df.columns)):
            # show the columns that are missing
            raise ValueError(
                "Data is missing required columns: {}".format(
                    set(expected_columns) - set(df.columns)
                )
            )

        # Check that the datetime index is timezone aware timestamp
        if not isinstance(df.index, pd.DatetimeIndex) and "datetime" not in df.columns:
            raise ValueError("Index is not datetime and datetime not provided")

        elif "datetime" in df.columns:
            if df["datetime"].dt.tz is None:
                raise ValueError("Datatime is missing timezone information")
            df["datetime"] = pd.to_datetime(df["datetime"])
            df.set_index("datetime", inplace=True)

        elif df.index.tz is None:
            raise ValueError("Datatime is missing timezone information")
        elif str(df.index.tz) == "UTC":
            self.warnings.append(
                EEMeterWarning(
                    qualified_name="eemeter.data_quality.utc_index",
                    description=(
                        "Datetime index is in UTC. Use tz_localize() with the local timezone to ensure correct aggregations"
                    ),
                    data={},
                )
            )
        self.tz = df.index.tz

        # prevent later issues when merging on generated datetimes, which default to ns precision
        # there is almost certainly a smoother way to accomplish this conversion, but this works
        if df.index.dtype.unit != "ns":
            utc_index = df.index.tz_convert("UTC")
            ns_index = utc_index.astype("datetime64[ns, UTC]")
            df.index = ns_index.tz_convert(self.tz)

        # Convert electricity data having 0 meter values to NaNs
        if self.is_electricity_data:
            df.loc[df["observed"] == 0, "observed"] = np.nan

        # Caltrack 2.3.2 - Drop duplicates
        df = remove_duplicates(df)

        df = self._get_contiguous_datetime(df)
        df = self._interpolate(df)
        df = self._add_pv_start_date(df)

        return df


class HourlyBaselineData(_HourlyData):
    """Data class to represent Hourly Baseline Data.

    Only baseline data should go into the dataframe input, no blackout data should be input.
    Checks sufficiency for the data provided as input depending on OpenEEMeter specifications and populates disqualifications and warnings based on it.

    Args:
        df (DataFrame): A dataframe having a datetime index or a datetime column with the timezone also being set.
            It also requires 2 more columns - 'observed' for meter data, and 'temperature' for temperature data.
            Optionally, column 'ghi' can be included in order to fit on solar data.
            The temperature column should have values in Fahrenheit. Please convert your temperatures accordingly.

        is_electricity_data (bool): Flag to ascertain if this is electricity data or not. Electricity data values of 0 are set to NaN.

    Attributes:
        df (DataFrame): Immutable dataframe that contains the meter and temperature values for the baseline data period.
        disqualification (list[EEMeterWarning]): A list of serious issues with the data that can degrade the quality of the model. If you want to go ahead with building the model while ignoring them, set the ignore_disqualification = True flag in the model. By default disqualifications are not ignored.
        warnings (list[EEMeterWarning]): A list of issues with the data, but none that will severely reduce the quality of the model built.
        pv_start (datetime.date): Solar install date. If left unset, assumed to be at beginning of data.

    """

    def _check_data_sufficiency(self):
        data = _create_sufficiency_df(self.df)
        hsc = HourlySufficiencyCriteria(
            data=data, is_electricity_data=self.is_electricity_data
        )
        hsc.check_sufficiency_baseline()
        disqualification = hsc.disqualification
        warnings = hsc.warnings

        return disqualification, warnings


class HourlyReportingData(_HourlyData):
    """Data class to represent Hourly Reporting Data.

    Only reporting data should go into the dataframe input, no blackout data should be input.
    Checks sufficiency for the data provided as input depending on OpenEEMeter specifications and populates disqualifications and warnings based on it.

    Meter data input is optional for the reporting class.

    Args:
        df (DataFrame): A dataframe having a datetime index or a datetime column with the timezone also being set.
            It also requires 2 more columns - 'observed' for meter data, and 'temperature' for temperature data.
            If GHI was provided during the baseline period, it should also be supplied for the reporting period with column name 'ghi'.
            The temperature column should have values in Fahrenheit. Please convert your temperatures accordingly.

        is_electricity_data (bool): Flag to ascertain if this is electricity data or not. Electricity data values of 0 are set to NaN.

    Attributes:
        df (DataFrame): Immutable dataframe that contains the meter and temperature values for the baseline data period.
        disqualification (list[EEMeterWarning]): A list of serious issues with the data that can degrade the quality of the model. If you want to go ahead with building the model while ignoring them, set the ignore_disqualification = True flag in the model. By default disqualifications are not ignored.
        warnings (list[EEMeterWarning]): A list of issues with the data, but none that will severely reduce the quality of the model built.
        pv_start (datetime.date): Solar install date. If left unset, assumed to be at beginning of data.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        is_electricity_data: bool,
        pv_start: Union[date, str, None] = None,
        **kwargs: dict,
    ):
        df = df.copy()
        if "observed" not in df.columns:
            df["observed"] = np.nan

        super().__init__(df, is_electricity_data, pv_start, **kwargs)

    def _check_data_sufficiency(self):
        data = _create_sufficiency_df(self.df)
        hsc = HourlySufficiencyCriteria(
            data=data, is_electricity_data=self.is_electricity_data
        )
        hsc.check_sufficiency_reporting()
        disqualification = hsc.disqualification
        warnings = hsc.warnings

        return disqualification, warnings


def _create_sufficiency_df(df: pd.DataFrame):
    """Creates dataframe equivalent to legacy hourly input"""
    df.loc[df["interpolated_observed"] == 1, "observed"] = np.nan
    df.loc[df["interpolated_temperature"] == 1, "temperature"] = np.nan
    if "ghi" in df.columns:
        df.loc[df["interpolated_ghi"] == 1, "ghi"] = np.nan
    # set temperature_not_null  to 1.0 if temperature is not null
    df["temperature_not_null"] = df["temperature"].notnull().astype(float)
    df["temperature_null"] = df["temperature"].isnull().astype(float)
    return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
# This setup.py inspired by https://github.com/kennethreitz/setup.py

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

NAME = "opendsm"
REQUIRED = [
    "attrs",
    "click",
    "fdasrsf>=2.4.1", # not sure if necessary anymore
    # "mkl-devel",  # needed for fdasrsf to work: https://github.com/jdtuck/fdasrsf_python/issues/41
    "nlopt",
    "numba",
    "numpy>=1.24.4",
    "pandas>=1.1.0",
    "pyarrow",    # why?
    "pydantic>=2.0",
    "pywavelets",
    "requests",
    "scikit-fda",
    "scikit-learn>=1.3.0",
    "scipy>=1.10.1",
    "statsmodels",
]

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = "\n" + f.read()

# Load the package's __version__.py module as a dictionary.
about = {}
with open(os.path.join(here, NAME, "__version__.py")) as f:
    exec(f.read(), about)


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPi via Twine…")
        os.system("twine upload dist/*")

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=about["__description__"],
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=about["__author__"],
    author_email=about["__author_email__"],
    url=about["__url__"],
    packages=find_packages(exclude=("tests",)),
    install_requires=REQUIRED,
    include_package_data=True,
    license=about["__license__"],
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    # $ setup.py publish support.
    cmdclass={"upload": UploadCommand},
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import numpy as np
from opendsm.eemeter.models.daily.utilities.selection_criteria import (
    neg_log_likelihood,
    selection_criteria,
)


def test_neg_log_likelihood():
    # Test case 1: Test with a simple loss and N
    loss = 1.0
    N = 10
    result = neg_log_likelihood(loss, N)
    expected = -2.6764598670764994
    assert np.allclose(result, expected)

    # Test case 2: Test with a larger loss and N
    loss = 10.0
    N = 100
    result = neg_log_likelihood(loss, N)
    expected = -26.764598670764993
    assert np.allclose(result, expected)

    # Test case 3: Test with a loss of zero (should return infinity)
    loss = 0.0
    N = 10
    result = neg_log_likelihood(loss, N)
    expected = np.inf
    assert np.allclose(result, expected)

    # Test case 4: Test with a negative loss (should raise ValueError)
    loss = -1.0
    N = 10
    try:
        neg_log_likelihood(loss, N)
    except ValueError as e:
        assert str(e) == "loss must be non-negative"

    # Test case 5: Test with a negative N (should raise ValueError)
    loss = 1.0
    N = -10
    try:
        neg_log_likelihood(loss, N)
    except ValueError as e:
        assert str(e) == "N must be positive"


def test_selection_criteria():
    # Test case 1: Test with RMSE criterion
    loss = 1.0
    TSS = 10.0
    N = 10
    num_coeffs = 2
    result = selection_criteria(
        loss, TSS, N, num_coeffs, model_selection_criteria="rmse"
    )
    expected = np.sqrt(loss / N)
    assert np.allclose(result, expected)

    # Test case 2: Test with RMSE adjusted criterion
    loss = 1.0
    TSS = 10.0
    N = 10
    num_coeffs = 2
    result = selection_criteria(
        loss, TSS, N, num_coeffs, model_selection_criteria="rmse_adj"
    )
    expected = np.sqrt(loss / (N - num_coeffs - 1))
    assert np.allclose(result, expected)

    # Test case 3: Test with R-squared criterion
    loss = 1.0
    TSS = 10.0
    N = 10
    num_coeffs = 2
    result = selection_criteria(
        loss, TSS, N, num_coeffs, model_selection_criteria="r_squared"
    )
    expected = (1 - (1 - loss / TSS)) * 10
    assert np.allclose(result, expected)

    # Test case 4: Test with adjusted R-squared criterion
    loss = 1.0
    TSS = 10.0
    N = 10
    num_coeffs = 2
    result = selection_criteria(
        loss, TSS, N, num_coeffs, model_selection_criteria="r_squared_adj"
    )
    expected = 1.2857142857142856
    assert np.allclose(result, expected)

    # Test case 5: Test with FPE criterion
    loss = 1.0
    TSS = 10.0
    N = 10
    num_coeffs = 2
    result = selection_criteria(
        loss, TSS, N, num_coeffs, model_selection_criteria="fpe"
    )
    expected = 0.18571428571428572
    assert np.allclose(result, expected)

    # Test case 6: Test with AIC criterion
    loss = 1.0
    TSS = 10.0
    N = 10
    num_coeffs = 2
    result = selection_criteria(
        loss, TSS, N, num_coeffs, model_selection_criteria="aic"
    )
    expected = 0.9352919734152998
    assert np.allclose(result, expected)

    # Test case 7: Test with AICc criterion
    loss = 1.0
    TSS = 10.0
    N = 10
    num_coeffs = 2
    result = selection_criteria(
        loss, TSS, N, num_coeffs, model_selection_criteria="aicc"
    )
    expected = 1.1067205448438713
    assert np.allclose(result, expected)

    # Test case 8: Test with CAIC criterion
    loss = 1.0
    TSS = 10.0
    N = 10
    num_coeffs = 2
    result = selection_criteria(
        loss, TSS, N, num_coeffs, model_selection_criteria="caic"
    )
    expected = 1.195808992014109
    assert np.allclose(result, expected)

    # Test case 9: Test with BIC criterion
    loss = 1.0
    TSS = 10.0
    N = 10
    num_coeffs = 2
    result = selection_criteria(
        loss, TSS, N, num_coeffs, model_selection_criteria="bic"
    )
    expected = 0.9958089920141091
    assert np.allclose(result, expected)

    # Test case 10: Test with SABIC criterion
    loss = 1.0
    TSS = 10.0
    N = 10
    num_coeffs = 2
    result = selection_criteria(
        loss, TSS, N, num_coeffs, model_selection_criteria="sabic"
    )
    expected = 0.3966625373033108
    assert np.allclose(result, expected)

    # Test case 11: Test with invalid criterion -> Not possible since we are using the settings config
    # loss = 1.0
    # TSS = 10.0
    # N = 10
    # num_coeffs = 2
    # try:
    #     selection_criteria(loss, TSS, N, num_coeffs, model_selection_criteria="invalid")
    # except ValueError as e:
    #     assert str(e) == "Invalid model selection criterion"

    # Remove the below test cases once the following criteria are implemented

    # Test case 12: Test with DIC criterion
    loss = 1.0
    TSS = 10.0
    N = 10
    num_coeffs = 2
    try:
        selection_criteria(loss, TSS, N, num_coeffs, model_selection_criteria="dic")
    except NotImplementedError as e:
        assert str(e) == "DIC has not been implmented as a model selection criterion"

    # Test case 13: Test with WAIC criterion
    loss = 1.0
    TSS = 10.0
    N = 10
    num_coeffs = 2
    try:
        selection_criteria(loss, TSS, N, num_coeffs, model_selection_criteria="waic")
    except NotImplementedError as e:
        assert str(e) == "WAIC has not been implmented as a model selection criterion"

    # Test case 14: Test with WBIC criterion
    loss = 1.0
    TSS = 10.0
    N = 10
    num_coeffs = 2
    try:
        selection_criteria(loss, TSS, N, num_coeffs, model_selection_criteria="wbic")
    except NotImplementedError as e:
        assert str(e) == "WBIC has not been implmented as a model selection criterion"
